# Copyright 2019 SuperDARN Canada, University of Saskatchewan
# Authors: Liam Graham, Marci Detwiller
"""
This file contains a utility class for restructuring Borealis files.
The files can be restructured from record-by-record format (as written on
site in real-time) to an array-based format where the number of records
is a dimension of the array. This makes the files much more human-readable
and also faster to read.

Classes
-------
BorealisRestructureUtilities

Exceptions
----------
BorealisRestructureError
"""
import datetime
import numpy as np

from collections import OrderedDict

from pydarn import borealis_exceptions, borealis_formats
from .borealis_utilities import BorealisUtilities


class BorealisRestructureUtilities():
    """
    Utility class containing static methods for restructuring Borealis files.

    Restructured files result in faster reads and a more human-readable
    format.

    The shared fields of the file are written as a single value in the file,
    and unshared fields are written as arrays where number of records is the
    first dimension.

    Static Methods
    --------------
    find_max_sequences(data)
        Find the max number of sequences between records in a site file, for
        restructuring to arrays.
    find_max_beams(data)
        Find the max number of beams between records in a site file, for
        restructuring to arrays.

    Class Methods
    -------------
    borealis_site_to_array_dict(origin_string, data_dict, conversion_type)
        uses the other BorealisRestructureUtilities
        methods to convert dictionary of records into a dictionary
        of arrays
    borealis_array_to_site_dict(origin_string, data_dict, conversion_type)
        uses the other BorealisRestructureUtilities
        methods to convert dictionary of arrays into a dictionary
        of records
    bfiq_site_to_array(data_dict, origin_string)
        Convert bfiq site data to array style dictionary.
    antennas_iq_site_to_array(data_dict, origin_string)
        Convert antennas_iq site data to array style dictionary.
    rawacf_site_to_array(data_dict, origin_string)
        Convert rawacf site data to array style dictionary.
    bfiq_array_to_site(data_dict, origin_string)
        Convert bfiq array data to site record dictionary.
    antennas_iq_array_to_site(data_dict, origin_string)
        Convert antennas_iq array data to site record dictionary.
    rawacf_array_to_site(data_dict, origin_string)
        Convert rawacf array data to site record dictionary.
    """

    @staticmethod
    def find_max_sequences(data: OrderedDict) -> int:
        """
        Finds the maximum number of sequences between records in a Borealis
        site style data file.

        Parameters
        ----------
        data
            Site formatted data from a Borealis file, organized as one
            record for each slice

        Returns
        -------
        max_seqs
            The largest number of sequences found in one record from the
            file
        """
        max_seqs = 0
        for k in data:
            if max_seqs < data[k]["num_sequences"]:
                max_seqs = data[k]["num_sequences"]
        return max_seqs

    @staticmethod
    def find_max_beams(data: OrderedDict) -> int:
        """
        Finds the maximum number of beams between records in a Borealis
        site style data file.

        Parameters
        ----------
        data
            Site formatted data from a Borealis file, organized as one
            record for each slice

        Returns
        -------
        max_beams
            The largest number of beams found in one record from the
            file
        """
        max_beams = 0
        for k in data:

            if max_beams < len(data[k]["beam_nums"]):
                max_beams = len(data[k]["beam_nums"])
        return max_beams

    @classmethod
    def _iq_site_to_array(cls, data_dict: OrderedDict,
                          iq_filetype: str) -> dict:
        """
        Base function for converting both bfiq and antennas_iq data to
        restructured array format.

        Parameters
        ----------
        data_dict: OrderedDict
            a dict of timestamped records loaded from an hdf5 Borealis bfiq
            or antennas_iq data file
        iq_filetype: str
            'bfiq' or 'antennas_iq' for more information on how to process.

        Returns
        -------
        new_data_dict
            A dictionary containing the data from data_dict
            reformatted to be stored entirely as arrays, or as
            one entry if the field does not change between records
        """
        new_data_dict = dict()
        num_records = len(data_dict)

        # write shared fields to dictionary
        first_key = list(data_dict.keys())[0]

        # find maximum number of sequences
        max_seqs = cls.find_max_sequences(data_dict)
        max_beams = cls.find_max_beams(data_dict)
        dims = data_dict[first_key]["data_dimensions"]

        if iq_filetype == 'antennas_iq':
            # dims are num_antennas, num_sequences, num_samps
            for field in borealis_formats.BorealisAntennasIq.shared_fields:
                new_data_dict[field] = data_dict[first_key][field]

            num_antennas = dims[0]
            num_samps = dims[2]
            data_buffer_offset = num_antennas * num_samps * max_seqs
            data_buffer = np.zeros(num_records * data_buffer_offset,
                                   dtype=np.complex64)

            data_descriptors = np.array(["num_records", "num_antennas",
                                         "max_num_sequences", "num_samps"],
                                        dtype=np.unicode_)
            data_shape = (num_records, num_antennas, max_seqs, num_samps)

        elif iq_filetype == 'bfiq':
            # dims are num_antenna_arrays, num_sequences, num_beams, num_samps
            for field in borealis_formats.BorealisBfiq.shared_fields:
                new_data_dict[field] = data_dict[first_key][field]
            num_arrays = dims[0]
            num_samps = dims[3]

            data_buffer_offset = num_arrays * max_beams * num_samps * max_seqs
            data_buffer = np.zeros(num_records * data_buffer_offset,
                                   dtype=np.complex64)
            data_descriptors = np.array(["num_records", "num_antenna_arrays",
                                         "max_num_sequences",
                                         "max_num_beams", "num_samps"],
                                        dtype=np.unicode_)
            data_shape = (num_records, num_arrays, max_seqs,
                          max_beams, num_samps)

        else:  # Unrecognized
            raise borealis_exceptions.BorealisRestructureError("Unrecognized"
                                                               " filetype in"
                                                               " _iq_site_to_array")

        sqn_buffer_offset = max_seqs
        sqn_ts_buffer = np.zeros(num_records * max_seqs)
        sqn_shape = (num_records, max_seqs)

        noise_buffer_offset = max_seqs
        noise_buffer = np.zeros(num_records * max_seqs)
        noise_shape = (num_records, max_seqs)

        sqn_num_array = np.empty((num_records), dtype=np.int64)
        int_time_array = np.empty((num_records), dtype=np.float32)
        scan_start_array = np.empty((num_records), dtype=np.bool_)
        num_slices_array = np.empty((num_records), dtype=np.int64)

        beam_nums_array = np.zeros((num_records, max_beams), dtype=np.uint32)
        beam_azms_array = np.zeros((num_records, max_beams), dtype=np.float64)
        num_beams_array = np.empty((num_records), dtype=np.uint32)

        for rec_idx, k in enumerate(data_dict.keys()):
            # handle unshared fields
            sqn_num_array[rec_idx] = data_dict[k]["num_sequences"]
            int_time_array[rec_idx] = data_dict[k]["int_time"]
            scan_start_array[rec_idx] = data_dict[k]["scan_start_marker"]
            num_slices_array[rec_idx] = data_dict[k]["num_slices"]

            beam_nums_buffer = data_dict[k]["beam_nums"]
            beam_azms_buffer = data_dict[k]["beam_azms"]
            # the beam_nums, beam_azms at a record may have appended zeros,
            # which is why a num_beams array is necessary.
            beam_nums_array[rec_idx][0:len(beam_nums_buffer)] =\
                beam_nums_buffer
            beam_azms_array[rec_idx][0:len(beam_azms_buffer)] =\
                beam_azms_buffer
            num_beams_array[rec_idx] = len(beam_nums_buffer)

            # insert data into buffer
            record_buffer = data_dict[k]["data"]
            data_pos = rec_idx * data_buffer_offset
            data_end = data_pos + len(record_buffer)
            data_buffer[data_pos:data_end] = record_buffer

            # insert sequence timestamps into buffer
            rec_sqn_ts = data_dict[k]["sqn_timestamps"]
            sqn_pos = rec_idx * sqn_buffer_offset
            sqn_end = sqn_pos + data_dict[k]["num_sequences"]
            sqn_ts_buffer[sqn_pos:sqn_end] = rec_sqn_ts

            rec_noise = data_dict[k]["noise_at_freq"]
            noise_pos = rec_idx * noise_buffer_offset
            noise_end = noise_pos + data_dict[k]["num_sequences"]
            noise_buffer[noise_pos:noise_end] = rec_noise

        # write leftover metadata and data
        new_data_dict["num_sequences"] = sqn_num_array
        new_data_dict["int_time"] = int_time_array
        new_data_dict["scan_start_marker"] = scan_start_array
        new_data_dict["num_slices"] = num_slices_array
        new_data_dict["beam_nums"] = beam_nums_array
        new_data_dict["beam_azms"] = beam_azms_array
        new_data_dict["num_beams"] = num_beams_array

        new_data_dict["data"] = data_buffer.reshape(data_shape)
        new_data_dict["sqn_timestamps"] = sqn_ts_buffer.reshape(sqn_shape)
        new_data_dict["noise_at_freq"] = noise_buffer.reshape(noise_shape)

        new_data_dict["data_descriptors"] = data_descriptors

        return new_data_dict

    @classmethod
    def bfiq_site_to_array(cls, data_dict: OrderedDict,
                           origin_string: str = '') -> dict:
        """
        Restructuring method for bfiq data.

        Parameters
        ----------
        origin_string
            String of where file originated from for better error messages.
        data_dict
            a dict of timestamped records loaded from an hdf5 Borealis bfiq
            data file

        Returns
        -------
        new_data_dict
            A dictionary containing the data from data_dict
            reformatted to be stored entirely as arrays, or as
            one entry if the field does not change between records
        """

        try:
            new_data_dict = cls._iq_site_to_array(data_dict, 'bfiq')

            BorealisUtilities.check_arrays(origin_string, new_data_dict,
                                           borealis_formats.BorealisBfiq.array_single_element_types(),
                                           borealis_formats.BorealisBfiq.array_array_dtypes(),
                                           borealis_formats.BorealisBfiq.unshared_fields)

        except Exception as e:
            raise borealis_exceptions.BorealisRestructureError('Error '
                                                               'restructuring'
                                                               ' bfiq from '
                                                               'site to array'
                                                               ' style:'
                                                               '{}'.format(e))\
                    from e

        return new_data_dict

    @classmethod
    def antennas_iq_site_to_array(cls, data_dict: OrderedDict,
                                  origin_string: str = '') -> dict:
        """
        Restructuring method for antennas_iq data.

        Parameters
        ----------
        origin_string
            String of where file originated from for better error messages.
        data_dict
            a dict of timestamped records loaded from an hdf5 Borealis
            antennas_iq data file

        Returns
        -------
        new_data_dict
            A dictionary containing the data from data_dict
            reformatted to be stored entirely as arrays, or as
            one entry if the field does not change between records
        """

        try:
            new_data_dict = cls._iq_site_to_array(data_dict, 'antennas_iq')

            BorealisUtilities.check_arrays(origin_string, new_data_dict,
                                           borealis_formats.BorealisAntennasIq.array_single_element_types(),
                                           borealis_formats.BorealisAntennasIq.array_array_dtypes(),
                                           borealis_formats.BorealisAntennasIq.unshared_fields)

        except Exception as e:
            raise borealis_exceptions.BorealisRestructureError('Error '
                                                               'restructuring'
                                                               ' antennas_iq '
                                                               'from site to'
                                                               ' array style:'
                                                               '{}'.format(e))\
                    from e

        return new_data_dict

    @classmethod
    def rawacf_site_to_array(cls, data_dict: OrderedDict,
                             origin_string: str = '') -> dict:
        """
        Restructuring method for rawacf data.

        Parameters
        ----------
        data_dict
            a dict of timestamped records loaded from an hdf5 Borealis rawacf
            data file
        origin_string
            String of where file originated from for better error messages.

        Returns
        -------
        new_data_dict
            A dictionary containing the data from data_dict
            reformatted to be stored entirely as arrays, or as
            one entry if the field does not change between records
        """
        try:
            new_data_dict = dict()
            num_records = len(data_dict)
            max_seqs = cls.find_max_sequences(data_dict)
            max_beams = cls.find_max_beams(data_dict)

            # write shared fields to dictionary
            first_key = list(data_dict.keys())[0]
            for field in borealis_formats.BorealisRawacf.shared_fields:
                new_data_dict[field] = data_dict[first_key][field]

            # handle unshared data fields
            first_dims = data_dict[first_key]["correlation_dimensions"]
            # dims are num_beams, num_ranges, num_lags
            num_ranges = first_dims[1]  # these don't change
            num_lags = first_dims[2]

            correlation_descriptors = np.array(['num_records', 'max_num_beams',
                                                'num_ranges', 'num_lags'],
                                               dtype=np.unicode_)
            correlation_shape = (num_records, max_beams, num_ranges, num_lags)

            noise_buffer_offset = max_seqs
            noise_buffer = np.zeros(num_records * max_seqs)
            noise_shape = (num_records, max_seqs)

            main_array = np.empty(correlation_shape, dtype=np.complex64)
            intf_array = np.empty(correlation_shape, dtype=np.complex64)
            xcfs_array = np.empty(correlation_shape, dtype=np.complex64)

            sqn_ts_array = np.empty((num_records, max_seqs), dtype=np.float64)
            sqn_num_array = np.empty((num_records), dtype=np.int64)
            int_time_array = np.empty((num_records), dtype=np.float32)
            scan_start_array = np.empty((num_records), dtype=np.bool_)
            num_slices_array = np.empty((num_records), dtype=np.int64)

            beam_nums_array = np.zeros((num_records, max_beams),
                                       dtype=np.uint32)
            beam_azms_array = np.zeros((num_records, max_beams),
                                       dtype=np.float64)
            num_beams_array = np.empty((num_records), dtype=np.uint32)

            for rec_idx, k in enumerate(data_dict.keys()):
                sqn_num_array[rec_idx] = data_dict[k]["num_sequences"]
                int_time_array[rec_idx] = data_dict[k]["int_time"]
                scan_start_array[rec_idx] = data_dict[k]["scan_start_marker"]

                scan_start_array[rec_idx] = data_dict[k]["scan_start_marker"]
                num_slices_array[rec_idx] = data_dict[k]["num_slices"]

                beam_nums_buffer = data_dict[k]["beam_nums"]
                beam_azms_buffer = data_dict[k]["beam_azms"]
                # the beam_nums, beam_azms at a record may have appended zeros,
                # which is why a num_beams array is necessary.
                beam_nums_array[rec_idx][0:len(beam_nums_buffer)] =\
                    beam_nums_buffer
                beam_azms_array[rec_idx][0:len(beam_azms_buffer)] =\
                    beam_azms_buffer
                num_beams = num_beams_array[rec_idx] = len(beam_nums_buffer)

                # some records have fewer than the specified
                # number of sequences get around this by zero
                # padding to the recorded number
                sqn_timestamps = data_dict[k]["sqn_timestamps"]
                while len(sqn_timestamps) < max_seqs:
                    sqn_timestamps = np.append(sqn_timestamps, 0)
                sqn_ts_array[rec_idx] = sqn_timestamps

                rec_noise = data_dict[k]["noise_at_freq"]
                noise_pos = rec_idx * noise_buffer_offset
                noise_end = noise_pos + data_dict[k]["num_sequences"]
                noise_buffer[noise_pos:noise_end] = rec_noise

                record_dims = data_dict[k]['correlation_dimensions']
                # anything greater than the number_of_beams will
                # remain as zeros
                main_array[rec_idx][:num_beams] =\
                    data_dict[k]["main_acfs"].reshape(record_dims)
                intf_array[rec_idx][:num_beams] =\
                    data_dict[k]["intf_acfs"].reshape(record_dims)
                xcfs_array[rec_idx][:num_beams] =\
                    data_dict[k]["xcfs"].reshape(record_dims)

            # write leftover metadata and data
            new_data_dict["num_sequences"] = sqn_num_array
            new_data_dict["int_time"] = int_time_array
            new_data_dict["scan_start_marker"] = scan_start_array
            new_data_dict["num_slices"] = num_slices_array
            new_data_dict["beam_nums"] = beam_nums_array
            new_data_dict["beam_azms"] = beam_azms_array
            new_data_dict["num_beams"] = num_beams_array
            new_data_dict["sqn_timestamps"] = sqn_ts_array
            new_data_dict["noise_at_freq"] = noise_buffer.reshape(noise_shape)

            new_data_dict["correlation_descriptors"] = correlation_descriptors
            new_data_dict["main_acfs"] = main_array
            new_data_dict["intf_acfs"] = intf_array
            new_data_dict["xcfs"] = xcfs_array

            BorealisUtilities.check_arrays(origin_string, new_data_dict,
                                           borealis_formats.BorealisRawacf.array_single_element_types(),
                                           borealis_formats.BorealisRawacf.array_array_dtypes(),
                                           borealis_formats.BorealisRawacf.unshared_fields)

        except Exception as e:
            raise borealis_exceptions.BorealisRestructureError('Error '
                                                               'restructuring'
                                                               ' rawacf from'
                                                               ' site to array'
                                                               ' style:'
                                                               '{}'.format(e))\
                    from e
        return new_data_dict

    @staticmethod
    def _iq_array_to_site(data_dict: dict, iq_filetype: str) -> OrderedDict:
        """
        Base function for converting both bfiq and antennas_iq back to
        original site format.

        Parameters
        ----------
        data_dict: dict
            An opened bfiq hdf5 file in array format
        iq_filetype: str
            'bfiq' or 'antennas_iq' to determine how to process

        Returns
        -------
        timestamp_dict
            A timestamped dictionary containing the data from data_dict
            formatted as the output from a site file (as records, where
            keys are timestamp of first sequence in the record)
        """

        timestamp_dict = OrderedDict()

        if iq_filetype == 'bfiq':
            # dims are num_antenna_arrays, num_sequences, num_beams, num_samps
            shared_fields = borealis_formats.BorealisBfiq.shared_fields
            unshared_fields = borealis_formats.BorealisBfiq.unshared_fields
            num_records, num_antenna_arrays, max_num_sequences, \
                max_num_beams, num_samps = data_dict["data"].shape
            # new data descriptors
            data_descriptors = np.array(["num_antenna_arrays",
                                         "num_sequences", "num_beams",
                                         "num_samps"],
                                        dtype=np.unicode_)
        elif iq_filetype == 'antennas_iq':
            # dims are num_antennas, num_sequences, num_samps
            shared_fields = borealis_formats.BorealisAntennasIq.shared_fields
            unshared_fields = \
                borealis_formats.BorealisAntennasIq.unshared_fields
            num_records, num_antennas, max_num_sequences, num_samps = \
                data_dict["data"].shape
            max_num_beams = max(data_dict["num_beams"])  # array only field
            # new data descriptors
            data_descriptors = np.array(["num_antennas", "num_sequences",
                                         "num_samps"], dtype=np.unicode_)
        else:
            raise borealis_exceptions.BorealisRestructureError("Unrecognized"
                                                               " filetype in"
                                                               " _iq_site_to_array")

        # get keys from first sequence timestamps
        for rec, seq_timestamp in enumerate(data_dict["sqn_timestamps"]):
            # format dictionary key in the same way it is done
            # in datawrite on site
            seq_datetime = datetime.datetime.utcfromtimestamp(seq_timestamp[0])
            epoch = datetime.datetime.utcfromtimestamp(0)
            key = str(int((seq_datetime - epoch).total_seconds() * 1000))

            timestamp_dict[key] = dict()
            # copy shared_fields
            for field in shared_fields:
                timestamp_dict[key][field] = data_dict[field]

            # data_descriptors needs to be overwritten
            timestamp_dict[key]["data_descriptors"] = data_descriptors

            # Handle per record fields
            for field in unshared_fields:
                timestamp_dict[key][field] = data_dict[field][rec]

            # Handle special cases - where dimensions may be
            # larger than necessary.
            num_beams = data_dict["num_beams"][rec]  # array only field
            if num_beams != max_num_beams:
                timestamp_dict[key]["beam_nums"] =\
                    timestamp_dict[key]["beam_nums"][rec, :num_beams]
                timestamp_dict[key]["beam_azms"] =\
                    timestamp_dict[key]["beam_azms"][rec, :num_beams]

            num_sequences = timestamp_dict[key]["num_sequences"]
            if num_sequences != max_num_sequences:
                timestamp_dict[key]["sqn_timestamps"] =\
                    data_dict["sqn_timestamps"][rec, :num_sequences]
                timestamp_dict[key]["noise_at_freq"] =\
                    data_dict["noise_at_freq"][rec, :num_sequences]

            # have to take correct dimensions to remove appended zeros if
            # num_sequences or
            # num_beams are less than their max.
            if iq_filetype == 'bfiq':
                dims = np.array([num_antenna_arrays, num_sequences,
                                 num_beams, num_samps],
                                dtype=np.uint32)
                new_data = data_dict["data"][rec, :dims[0], :dims[1],
                                             :dims[2], :dims[3]]
            elif iq_filetype == 'antennas_iq':
                dims = np.array([num_antennas, num_sequences, num_samps],
                                dtype=np.uint32)
                new_data = data_dict["data"][rec, :dims[0], :dims[1], :dims[2]]

            timestamp_dict[key]["data_dimensions"] = dims  # site only field
            timestamp_dict[key]["data"] = new_data.flatten()

        return timestamp_dict

    @classmethod
    def bfiq_array_to_site(cls, data_dict: dict,
                           origin_string: str = '') -> OrderedDict:
        """
        Converts a restructured array bfiq file back to the original site
        format.

        Parameters
        ----------
        origin_string
            String of where file originated from for better error messages.
        data_dict
            An opened bfiq hdf5 file in array format

        Returns
        -------
        timestamp_dict
            A timestamped dictionary containing the data from data_dict
            formatted as the output from a site file (as records, where
            keys are timestamp of first sequence in the record)
        """
        try:
            timestamp_dict = cls._iq_array_to_site(data_dict, 'bfiq')
            BorealisUtilities.check_records(origin_string, timestamp_dict,
                                            borealis_formats.BorealisBfiq.site_single_element_types(),
                                            borealis_formats.BorealisBfiq.site_array_dtypes())

        except Exception as e:
            raise borealis_exceptions.BorealisRestructureError('Error '
                                                               'restructuring'
                                                               ' bfiq from'
                                                               ' array to'
                                                               ' site style:'
                                                               '{}'.format(e))\
                    from e
        return timestamp_dict

    @classmethod
    def antennas_iq_array_to_site(cls, data_dict: dict,
                                  origin_string: str = '') -> OrderedDict:
        """
        Converts a restructured array antennas_iq file back to the
        original site format.

        Parameters
        ----------
        origin_string
            String of where file originated from for better error messages.
        data_dict
            An opened antennas_iq hdf5 file in array format

        Returns
        -------
        timestamp_dict
            A timestamped dictionary containing the data from data_dict
            formatted as the output from a site file (as records, where
            keys are timestamp of first sequence in the record)
        """
        try:
            timestamp_dict = cls._iq_array_to_site(data_dict, 'antennas_iq')
            BorealisUtilities.check_records(origin_string, timestamp_dict,
                                            borealis_formats.BorealisAntennasIq.site_single_element_types(),
                                            borealis_formats.BorealisAntennasIq.site_array_dtypes())

        except Exception as e:
            raise borealis_exceptions.BorealisRestructureError('Error '
                                                               'restructuring'
                                                               ' antennas_iq '
                                                               'from array to'
                                                               ' site style:'
                                                               '{}'.format(e))\
                    from e
        return timestamp_dict

    @staticmethod
    def rawacf_array_to_site(data_dict: dict,
                             origin_string: str = '') -> OrderedDict:
        """
        Converts a restructured array rawacf file back to the
        original site format.

        Parameters
        ----------
        data_dict
            An opened rawacf hdf5 file in array format
        origin_string
            String of where file originated from for better error messages.

        Returns
        -------
        timestamp_dict
            A timestamped dictionary containing the data from data_dict
            formatted as the output from a site file (as records, where
            keys are timestamp of first sequence in the record)
        """
        shared_fields = borealis_formats.BorealisRawacf.shared_fields
        unshared_fields = borealis_formats.BorealisRawacf.unshared_fields

        try:
            (num_records, max_num_beams, num_ranges, num_lags) = \
                data_dict["main_acfs"].shape
            # new correlation descriptors
            correlation_descriptors = np.array(['num_beams', 'num_ranges',
                                                'num_lags'],
                                               dtype=np.unicode_)
            # array only field
            max_num_sequences = max(data_dict["num_sequences"])

            timestamp_dict = OrderedDict()
            # get keys from first sequence timestamps
            for rec, seq_timestamp in enumerate(data_dict["sqn_timestamps"]):
                # format dictionary key in the same way it is done
                # in datawrite on site
                seq_datetime =\
                    datetime.datetime.utcfromtimestamp(seq_timestamp[0])
                epoch = datetime.datetime.utcfromtimestamp(0)
                key = str(int((seq_datetime - epoch).total_seconds() * 1000))

                timestamp_dict[key] = dict()
                for field in shared_fields:
                    timestamp_dict[key][field] = data_dict[field]
                # overwrite the correlation descriptors
                timestamp_dict[key]["correlation_descriptors"] =\
                    correlation_descriptors

                # Handle per record fields, copy all and deal with
                # dimensions issues after
                for field in unshared_fields:
                    timestamp_dict[key][field] = data_dict[field][rec]

                # Handle special cases - where dimensions may be
                # larger than necessary.
                num_beams = data_dict["num_beams"][rec]  # array only field
                if num_beams != max_num_beams:
                    timestamp_dict[key]["beam_nums"] =\
                        timestamp_dict[key]["beam_nums"][rec, :num_beams]
                    timestamp_dict[key]["beam_azms"] =\
                        timestamp_dict[key]["beam_azms"][rec, :num_beams]

                num_sequences = timestamp_dict[key]["num_sequences"]
                if num_sequences != max_num_sequences:
                    timestamp_dict[key]["sqn_timestamps"] =\
                        data_dict["sqn_timestamps"][rec, :num_sequences]
                    timestamp_dict[key]["noise_at_freq"] =\
                        data_dict["noise_at_freq"][rec, :num_sequences]

                dims = np.array([num_beams, num_ranges, num_lags],
                                dtype=np.uint32)
                main_acfs = np.copy(data_dict["main_acfs"][rec, :dims[0],
                                                           :dims[1], :dims[2]])
                intf_acfs = np.copy(data_dict["intf_acfs"][rec, :dims[0],
                                                           :dims[1], :dims[2]])
                xcfs = np.copy(data_dict["xcfs"][rec, :dims[0], :dims[1],
                                                 :dims[2]])

                timestamp_dict[key]["main_acfs"] = main_acfs.flatten()
                timestamp_dict[key]["intf_acfs"] = intf_acfs.flatten()
                timestamp_dict[key]["xcfs"] = xcfs.flatten()
                # site only field
                timestamp_dict[key]["correlation_dimensions"] = dims

            BorealisUtilities.check_records(origin_string, timestamp_dict,
                                            borealis_formats.BorealisRawacf.site_single_element_types(),
                                            borealis_formats.BorealisRawacf.site_array_dtypes())

        except Exception as e:
            raise borealis_exceptions.BorealisRestructureError('Error '
                                                               'restructuring'
                                                               ' rawacf from'
                                                               ' array to '
                                                               'site style:'
                                                               '{}'.format(e))\
                    from e
        return timestamp_dict

    @classmethod
    def borealis_site_to_array_dict(cls, origin_string: str,
                                    data_dict: OrderedDict,
                                    conversion_type: str) -> dict:
        """
        Converts a file from site style to restructured array style. Determines
        which base function to call based on conversion_type.

        Parameters
        ----------
        origin_string: str
            Filename or origin string for better error messages.
        data_dict: OrderedDict
            An opened rawacf hdf5 file in site record-by-record format
        conversion_type: str
            'bfiq', 'antennas_iq' or 'rawacf' to determine keys to convert

        Returns
        -------
        new_dict
            A dictionary containing the data from data_dict
            formatted to the array format
        """
        if conversion_type == 'bfiq':
            new_dict = cls.bfiq_site_to_array(data_dict, origin_string)
        elif conversion_type == 'rawacf':
            new_dict = cls.rawacf_site_to_array(data_dict, origin_string)
        elif conversion_type == 'antennas_iq':
            new_dict = cls.antennas_iq_site_to_array(data_dict, origin_string)
        else:
            raise borealis_exceptions.BorealisRestructureError('File type {} '
                                                               'not recognized'
                                                               'as '
                                                               'restructureable'
                                                               ' from site to '
                                                               'array style'
                                                               ''.format(conversion_type))
        return new_dict

    @classmethod
    def borealis_array_to_site_dict(cls, origin_string: str, data_dict: dict,
                                    conversion_type: str) -> OrderedDict:
        """
        Converts a file back to its original site format. Determines
        which base function to call based on conversion_type.

        Parameters
        ----------
        origin_string: str
            Filename or origin string for better error messages.
        data_dict: dict
            An opened rawacf hdf5 file in array format
        conversion_type: str
            'bfiq', 'antennas_iq' or 'rawacf' to determine keys to convert

        Returns
        -------
        new_dict
            A timestamped dictionary containing the data from data_dict
            formatted as the output from a site file.
        """
        if conversion_type == 'bfiq':
            new_dict = cls.bfiq_array_to_site(data_dict, origin_string)
        elif conversion_type == 'rawacf':
            new_dict = cls.rawacf_array_to_site(data_dict, origin_string)
        elif conversion_type == 'antennas_iq':
            new_dict = cls.antennas_iq_array_to_site(data_dict, origin_string)
        else:
            raise borealis_exceptions.BorealisRestructureError('File type {} '
                                                               'not recognized'
                                                               ' as '
                                                               'restructureable'
                                                               ' from array to'
                                                               ' site style'
                                                               ''.format(conversion_type))
        return new_dict



# Copyright 2018 SuperDARN
# Authors: Marina Schmidt
"""
This file contains several class pertaining to dmap format types
that are used by SuperDARN.

For description of these fields and their units see the following reference
materials:
    https://superdarn.github.io/rst/superdarn/src.doc/rfc/index.html
    https://radar-software-toolkit-rst.readthedocs.io/en/latest/
"""


class Rawacf():
    """
    Class containing Rawacf fields
    """
    types = {
        'radar.revision.major': 'c',
        'radar.revision.minor': 'c',
        'origin.code': 'c',
        'origin.time': 's',
        'origin.command': 's',
        'cp': 'h',
        'stid': 'h',
        'time.yr': 'h',
        'time.mo': 'h',
        'time.dy': 'h',
        'time.hr': 'h',
        'time.mt': 'h',
        'time.sc': 'h',
        'time.us': 'i',
        'txpow': 'h',
        'nave': 'h',
        'atten': 'h',
        'lagfr': 'h',
        'smsep': 'h',
        'ercod': 'h',
        'stat.agc': 'h',
        'stat.lopwr': 'h',
        'noise.search': 'f',
        'noise.mean': 'f',
        'channel': 'h',
        'bmnum': 'h',
        'bmazm': 'f',
        'scan': 'h',
        'offset': 'h',
        'rxrise': 'h',
        'intt.sc': 'h',
        'intt.us': 'i',
        'txpl': 'h',
        'mpinc': 'h',
        'mppul': 'h',
        'mplgs': 'h',
        'nrang': 'h',
        'frang': 'h',
        'rsep': 'h',
        'xcf': 'h',
        'tfreq': 'h',
        'mxpwr': 'i',
        'lvmax': 'i',
        'rawacf.revision.major': 'i',
        'rawacf.revision.minor': 'i',
        'combf': 's',
        'thr': 'f',
        'ptab': 'h',
        'ltab': 'h',
        'slist': 'h',
        'pwr0': 'f',
        'acfd': 'f',
        }
    extra_fields = {
            'ifmode': 'h',
            'mplgexs': 'h',
            }
    cross_correlation_field = {
            'xcfd': 'f'
            }


class Fitacf():
    """
    Class containing Fitacf fields
    """
    # Standard fields
    types = {
        'radar.revision.major': 'c',
        'radar.revision.minor': 'c',
        'origin.code': 'c',
        'origin.time': 's',
        'origin.command': 's',
        'cp': 'h',
        'stid': 'h',
        'time.yr': 'h',
        'time.mo': 'h',
        'time.dy': 'h',
        'time.hr': 'h',
        'time.mt': 'h',
        'time.sc': 'h',
        'time.us': 'i',
        'txpow': 'h',
        'nave': 'h',
        'atten': 'h',
        'lagfr': 'h',
        'smsep': 'h',
        'ercod': 'h',
        'stat.agc': 'h',
        'stat.lopwr': 'h',
        'noise.search': 'f',
        'noise.mean': 'f',
        'channel': 'h',
        'bmnum': 'h',
        'bmazm': 'f',
        'scan': 'h',
        'offset': 'h',
        'rxrise': 'h',
        'intt.sc': 'h',
        'intt.us': 'i',
        'txpl': 'h',
        'mpinc': 'h',
        'mppul': 'h',
        'mplgs': 'h',
        'nrang': 'h',
        'frang': 'h',
        'rsep': 'h',
        'xcf': 'h',
        'tfreq': 'h',
        'mxpwr': 'i',
        'lvmax': 'i',
        'combf': 's',
        'fitacf.revision.major': 'i',
        'fitacf.revision.minor': 'i',
        'noise.sky': 'f',
        'noise.lag0': 'f',
        'noise.vel': 'f',
        'ptab': 'h',
        'ltab': 'h',
        'pwr0': 'f'}
    extra_fields = {
            'ifmode': 'h',
            'mplgexs': 'h',
            }
    # Fields added if the data is good and can be
    # fitted
    fitted_fields = {
        'slist': 'h',
        'nlag': 'h',
        'qflg': 'c',
        'gflg': 'c',
        'p_l': 'f',
        'p_l_e': 'f',
        'p_s': 'f',
        'p_s_e': 'f',
        'v': 'f',
        'v_e': 'f',
        'w_l': 'f',
        'w_l_e': 'f',
        'w_s': 'f',
        'w_s_e': 'f',
        'sd_l': 'f',
        'sd_s': 'f',
        'sd_phi': 'f'}

    elevation_fields = {
        'x_qflg': 'c',
        'x_gflg': 'c',
        'x_p_l': 'f',
        'x_p_l_e': 'f',
        'x_p_s': 'f',
        'x_p_s_e': 'f',
        'x_v': 'f',
        'x_v_e': 'f',
        'x_w_l': 'f',
        'x_w_l_e': 'f',
        'x_w_s': 'f',
        'x_w_s_e': 'f',
        'phi0': 'f',
        'phi0_e': 'f',
        'elv': 'f',
        'elv_low': 'f',
        'elv_high': 'f',
        'x_sd_l': 'f',
        'x_sd_s': 'f',
        'x_sd_phi': 'f'
    }


class Grid():
    """
    Class containing the Grid fields
    """
    # Standard fields
    types = {
        'start.year': 'h',
        'start.month': 'h',
        'start.day': 'h',
        'start.hour': 'h',
        'start.minute': 'h',
        'start.second': 'd',
        'end.year': 'h',
        'end.month': 'h',
        'end.day': 'h',
        'end.hour': 'h',
        'end.minute': 'h',
        'end.second': 'd',
        'stid': 'h',
        'channel': 'h',
        'nvec': 'h',
        'freq': 'f',
        'major.revision': 'h',
        'minor.revision': 'h',
        'program.id': 'h',
        'noise.mean': 'f',
        'noise.sd': 'f',
        'gsct': 'h',
        'v.min': 'f',
        'v.max': 'f',
        'p.min': 'f',
        'p.max': 'f',
        'w.min': 'f',
        'w.max': 'f',
        've.min': 'f',
        've.max': 'f'}
    # Fields added if the data is good and was fitted
    # in the generation process of fitacf
    fitted_fields = {'vector.mlat': 'f',
                     'vector.vel.median': 'f',
                     'vector.channel': 'h',
                     'vector.stid': 'h',
                     'vector.vel.sd': 'f',
                     'vector.index': 'i',
                     'vector.kvect': 'f',
                     'vector.mlon': 'f'}
    # Extra fields if you include -ext in grid generation process
    extra_fields = {
        'vector.pwr.median': 'f',
        'vector.pwr.sd': 'f',
        'vector.wdt.median': 'f',
        'vector.wdt.sd': 'f'
    }


class Map():
    """
    Class containing possible Map fields
    """
    # Standard fields
    types = {
        'start.year': 'h',
        'start.month': 'h',
        'start.day': 'h',
        'start.hour': 'h',
        'start.minute': 'h',
        'start.second': 'd',
        'end.year': 'h',
        'end.month': 'h',
        'end.day': 'h',
        'end.hour': 'h',
        'end.minute': 'h',
        'end.second': 'd',
        'map.major.revision': 'h',
        'map.minor.revision': 'h',
        'doping.level': 'h',
        'model.wt': 'h',
        'error.wt': 'h',
        'IMF.flag': 'h',
        'IMF.delay': 'h',
        'IMF.Bx': 'd',
        'IMF.By': 'd',
        'IMF.Bz': 'd',
        'IMF.Vx': 'd',
        'IMF.tilt': 'd',
        'IMF.Kp': 'd',
        'hemisphere': 'h',
        'noigrf': 'h',
        'fit.order': 'h',
        'latmin': 'f',
        'chi.sqr': 'd',
        'chi.sqr.dat': 'd',
        'rms.err': 'd',
        'lon.shft': 'f',
        'lat.shft': 'f',
        'mlt.start': 'd',
        'mlt.end': 'd',
        'mlt.av': 'd',
        'pot.drop': 'd',
        'pot.drop.err': 'd',
        'pot.max': 'd',
        'pot.max.err': 'd',
        'pot.min': 'd',
        'pot.min.err': 'd',
        'stid': 'h',
        'channel': 'h',
        'nvec': 'h',
        'freq': 'f',
        'major.revision': 'h',
        'minor.revision': 'h',
        'program.id': 'h',
        'noise.mean': 'f',
        'noise.sd': 'f',
        'gsct': 'h',
        'v.min': 'f',
        'v.max': 'f',
        'p.min': 'f',
        'p.max': 'f',
        'w.min': 'f',
        'w.max': 'f',
        've.min': 'f',
        've.max': 'f',
        'vector.mlat': 'f',
        'vector.mlon': 'f',
        'vector.kvect': 'f',
        'vector.stid': 'h',
        'vector.channel': 'h',
        'vector.index': 'i',
        'vector.vel.median': 'f',
        'vector.vel.sd': 'f',
        }
    # Additional fields if you use map_addhmb
    hmb_fields = {
        'model.mlat': 'f',
        'model.mlon': 'f',
        'model.kvect': 'f',
        'model.vel.median': 'f',
        'boundary.mlat': 'f',
        'boundary.mlon': 'f',
    }
    # Additional fields if you use map_addmodel
    model_fields = {
         'model.angle': 's',
         'model.level': 's',
         'model.tilt': 's',
         'model.name': 's',
    }
    # Additional fields if you use map_addfit
    fit_fields = {
        'source': 's',
        'N': 'd',
        'N+1': 'd',
        'N+2': 'd',
        'N+3': 'd',
    }
    # Additional fields if you add -ext when
    # creating the map file in RST
    extra_fields = {
        'vector.pwr.median': 'f',
        'vector.pwr.sd': 'f',
        'vector.wdt.median': 'f',
        'vector.wdt.sd': 'f',
    }


class Iqdat():
    """
    Class contains Iqdat fields
    """
    # Standard Iqdat fields
    types = {
        'radar.revision.major': 'c',
        'radar.revision.minor': 'c',
        'origin.code': 'c',
        'origin.time': 's',
        'origin.command': 's',
        'cp': 'h',
        'stid': 'h',
        'time.yr': 'h',
        'time.mo': 'h',
        'time.dy': 'h',
        'time.hr': 'h',
        'time.mt': 'h',
        'time.sc': 'h',
        'time.us': 'i',
        'txpow': 'h',
        'nave': 'h',
        'atten': 'h',
        'lagfr': 'h',
        'smsep': 'h',
        'ercod': 'h',
        'stat.agc': 'h',
        'stat.lopwr': 'h',
        'noise.search': 'f',
        'noise.mean': 'f',
        'channel': 'h',
        'bmnum': 'h',
        'bmazm': 'f',
        'scan': 'h',
        'offset': 'h',
        'rxrise': 'h',
        'intt.sc': 'h',
        'intt.us': 'i',
        'txpl': 'h',
        'mpinc': 'h',
        'mppul': 'h',
        'mplgs': 'h',
        'nrang': 'h',
        'frang': 'h',
        'rsep': 'h',
        'xcf': 'h',
        'tfreq': 'h',
        'mxpwr': 'i',
        'lvmax': 'i',
        'iqdata.revision.major': 'i',
        'iqdata.revision.minor': 'i',
        'combf': 's',
        'seqnum': 'i',
        'chnnum': 'i',
        'smpnum': 'i',
        'skpnum': 'i',
        'ptab': 'h',
        'ltab': 'h',
        'tsc': 'i',
        'tus': 'i',
        'tatten': 'h',
        'tnoise': 'f',
        'toff': 'i',
        'tsze': 'i',
        'data': 'h',
        }



# Copyright 2018 SuperDARN Canada, University of Saskatchewan
# Authors: Marina Schmidt and Keith Kotyk
"""
This file contains classes to reading and writing of  formats used by
SuperDARN.

Classes:
--------
DmapRead : Reads s
DmapWrite : writes DMap record structure into a

Exceptions:
-----------
EmptyFileError
CursorError
DmapDataError
DmapDataTypeError
ZeroByteError
MismatchByteError
ValueError

Future work
-----------
Organization
    rethink public and private methods? <--- discussion

Notes
-----
DmapRead and DmapWrite are updated and improved versions of
backscatter's pydmap:
git@github.com:SuperDARNCanada/backscatter.git
written by Keith Kotyk
"""

import collections
import logging
import numpy as np
import os
import struct

from typing import List, Union

from pydarn import dmap_exceptions, DmapArray, DmapScalar, dmap2dict, dict2dmap

# Keeping these global definitions for readability purposes
# Data types use in s
DMAP = 0
# CHAR is defined as an int8 in RST rtypes.h https://github.com/SuperDARN/rst
CHAR = 1
SHORT = 2
INT = 3
FLOAT = 4
DOUBLE = 8
STRING = 9
LONG = 10
UCHAR = 16
USHORT = 17
UINT = 18
ULONG = 19

# Dictionary of DMap types (key) to quickly convert the format and byte size
# (value-tuple)
# DMap stands for DataMap documented in:
# https://superdarn.github.io/rst/superdarn/src.doc/rfc/0006.html
# https://radar-software-toolkit-rst.readthedocs.io/en/latest/
DMAP_DATA_TYPES = {DMAP: ('', 0),
                   CHAR: ('c', 1),  # CHAR is defined as an int8 in RST
                   SHORT: ('h', 2),  # defined as int16
                   INT: ('i', 4),  # defined as int32
                   FLOAT: ('f', 4),  # defined as float32
                   DOUBLE: ('d', 8),  # equavalece to float64
                   STRING: ('s', 1),  # strings - array of 1 byte characters
                   LONG: ('q', 8),  # defined as int64
                   UCHAR: ('B', 1),  # unsigned int8
                   USHORT: ('H', 2),  # unsigned int16
                   UINT: ('I', 4),  # unsigned int32
                   ULONG: ('Q', 8)}  # unsigned int64
# unsigned refers to how the integer is stored in the byte string, meaning
# the first bit is part of the number and not referring
# to the sign of the integer

pydarn_log = logging.getLogger('pydarn')


class DmapRead():
    """
    Reading and testing the integrity of s/stream.
    DMap is describe in the RST documentation:
    - https://superdarn.github.io/rst/superdarn/src.doc/rfc/0006.html
    - https://radar-software-toolkit-rst.readthedocs.io/en/latest/
    ...

    Attributes
    ----------
    dmap_file : str
         name or data stream (give data_stream=True)
    cursor : int
        Current position in the byte array
    dmap_end_bytes : int
        The length of the byte array

    Methods
    --------
    test_initial_data_integrity()
        Quickly reads the byte array for any errors
    read_records()
        Reads the byte array to obtain the DMap records
    zero_negative_check(element, element_name)
        Checks if the element is equal to zero
    check_data_type(data_type, data_name)
        Checks if the data_type exists in DMAP_DATA_TYPE dictionary
    read_record()
        Reads a record from the byte array
    read_scalar()
        Reads a scalar from the byte array
    read_array(record_size)
        Reads an array from the byte array
    read_string_DMAP_array(Shape, array_type)
        Reads an array of strings/DMap types into an array
    read_data()
        Reads the data type from the byte array
    read_numerical_array(data_type, shape, total_number_cells)
        Reads an array into a numpy array
    """

    def __init__(self, dmap_file: Union[str, bytes], data_stream=False):
        """
        Reads the /stream into a byte array for further reading of the
        DMap records.

        Parameters
        ----------
        dmap_file : str or bytes
                    file name or data stream (given data_stream=True)
                    containing dmap data.
        data_stream : bool
                 default to false, boolean that indicates if dmap_file is a
                 data stream
        Raises
        ------
        EmptyFileError
            dmap_file is empty
        FileNotFoundError
            dmap_file path does not exist

        See Also
        --------
        read_records : to obtain dmap_records
        """
        self.rec_num = 0
        self.cursor = 0  # Current position in bytes
        self.dmap_end_bytes = 0  # total number of bytes in the dmap_file

        """
        DMap records are stored in a deque data structure for
        memory efficiency and performance. Acts the same as a stack/list.
        See DEVELOPER_README.md for more information.
        """

        # _dmap_records are private to avoid tampering.
        self._dmap_records = collections.deque()
        self._records = collections.deque()
        self.dmap_file = dmap_file

        # read the whole file/stream into a byte array
        if data_stream is False:
            # check if the file is empty
            if os.path.getsize(self.dmap_file) == 0:
                raise dmap_exceptions.EmptyFileError(self.dmap_file)

            # Read binary
            with open(self.dmap_file, 'rb') as f:
                self.dmap_bytearr = bytearray(f.read())
            pydarn_log.debug("DMap Read file: {}".format(self.dmap_file))

        else:
            if len(dmap_file) == 0:
                raise dmap_exceptions.EmptyFileError("data stream")

            self.dmap_bytearr = bytearray(self.dmap_file)
            self.dmap_file = "stream"
            pydarn_log.debug("DMap Read file: Stream")
        self.dmap_buffer = memoryview(self.dmap_bytearr)
        self.dmap_end_bytes = len(self.dmap_bytearr)
        if self.dmap_end_bytes == 0:
            raise dmap_exceptions.EmptyFileError(self.dmap_file)

    def __repr__(self):
        """ for representation of the class object"""
        # __class__.__name__ allows to grab the class name such that
        # when a class inherits this one, the class name will be the child
        # class and not the parent class (dmap classes)
        return "{class_name}({filename}, {cursor}, {rec_num}, {total})"\
               "".format(class_name=self.__class__.__name__,
                         filename=self.dmap_file,
                         cursor=self.cursor,
                         total=self.dmap_end_bytes,
                         rec_num=self.rec_num)

    def __str__(self):
        """ for printing of the class object"""
        # __class__.__name__ allows to grab the class name such that
        # when a class inherits this one, the class name will be the child
        # class and not the parent class (dmap classes)
        return "Reading from {filename} at cursor: {cursor} "\
               "record number: {rec_num} with"\
               " a total number of bytes: {total_bytes}"\
               "".format(filename=self.dmap_file,
                         cursor=self.cursor,
                         rec_num=self.rec_num,
                         total_bytes=self.dmap_end_bytes)

    @property
    def get_dmap_records(self):
        """
        Returns the DMap structured records
        """
        return self._dmap_records

    @property
    def get_records(self):
        """
        Returns records of fields in a dictionary format
        """
        return self._records

    def zero_negative_check(self, element: int, element_name: str):
        """
        Checks if the element <= 0 bytes. If true then raise a ZeroByteError.

        Parameter
        ---------
        element : int
            An element to check if it is zero
        element_name : str
            The element's name title for a more detailed exception message

        Raises
        ------
        ZeroByteError
            if element <= 0
        """

        if element == 0:
            element_info = "{name} {size}".format(name=element_name,
                                                  size=element)
            raise dmap_exceptions.ZeroByteError(self.dmap_file, element_info,
                                                self.rec_num)
        elif element < 0:
            element_info = "{name} {size}".format(name=element_name,
                                                  size=element)
            raise dmap_exceptions.NegativeByteError(self.dmap_file,
                                                    element_info,
                                                    self.rec_num)

    def bytes_check(self, element: int, element_name: str, byte_check: int,
                    byte_check_name: str):
        """
        Checks if the element greater than the value comparing against
        (byte_check).

        Parameter
        ---------
        element : int
            An element to check if it is greater than another value in bytes
        element_name : str
            Name of the element for a more detailed exception message
        byte_check : int
            The byte value for the element to compare against
        byte_check_name : str
            Name of the byte value to be compared too for a more detailed
            exception message

        Raises
        ------
        MismatchByteError
             if element > byte_check
        """
        if element > byte_check:
            element_info = "{name} {size} >"\
                " {check_name} {check}".format(name=element_name,
                                               size=element,
                                               check_name=byte_check_name,
                                               check=byte_check)
            raise dmap_exceptions.MismatchByteError(self.dmap_file,
                                                    element_info,
                                                    self.rec_num)

    def check_data_type(self, data_type: int, data_name: str):
        """
        Check if the data_type is a DMap data type.

        Parameter
        ----------
        data_type : int
            Data_type read in by the byte array
        data_name : str
            Name of the data trying to read in

        Raises
        -------
        DmapDataTypeError
            if the data_type is not in the DMAP_DATA_TYPES dictionary
        """
        if data_type not in DMAP_DATA_TYPES.keys():
            raise dmap_exceptions.DmapDataTypeError(self.dmap_file,
                                                    data_name,
                                                    data_type,
                                                    self.rec_num)

    def test_initial_data_integrity(self):
        """
        Quick method for testing the integrity of the dmap data.

        Raises
        ------
        CursorError
            If the cursor is not set to an expected value.
        DmapDataError
            If the data is corrupted by some byte offset.

        See Also
        --------
        zero_check : raises ZeroByteError
        byte_check : raises MistmatchByteError
        """
        pydarn_log.debug("Testing the integrity of the /stream")
        total_block_size = 0  # unit of bytes
        if self.cursor != 0:
            raise dmap_exceptions.CursorError(self.cursor, 0, self.rec_num)

        while self.cursor < self.dmap_end_bytes:
            """
            s headers contain the following:
                - encoding identifier: is a unique 32-bit integer that
                  indicates how the block was constructed.
                  It is used to differentiate between the possible future
                  changes to the DataMap format.
                - block size: 32-bit integer that represents the total size of
                  block including the header and the data.
                - Number of Scalars: number of scalar variables
                - Scalar data: the scalar data of the record.
                  Please see DmapScalar for more information on scalars.
                - Number of arrays: number of array variables
                  Please see DmapArray for more information on arrays.
            """
            # This is an unused variable but is need to move the cursor to the
            # next offset.
            # TODO: Possible check that uses the encoding identifier
            # encoding_identifier = self.read_data('i',4)
            self.cursor += 4
            block_size = self.read_data('i', 4)
            self.zero_negative_check(block_size, "block size")

            total_block_size += block_size
            self.bytes_check(total_block_size, "total block size",
                             self.dmap_end_bytes, "total bytes in file")

            # 2 is to include the encoding_identifier and size of data which
            # are both int types.
            self.cursor = self.cursor + block_size - 2 *\
                DMAP_DATA_TYPES[INT][1]

        if total_block_size != self.dmap_end_bytes:
            message = "Error: Initial integrity check shows"\
                " total block size: {total_size} < end bytes {end_bytes}."\
                " Failed at record {rec}."\
                "".format(total_size=total_block_size,
                          end_bytes=self.dmap_end_bytes,
                          rec=self.rec_num)
            raise dmap_exceptions.DmapDataError(self.data_file, message)
        self.cursor = 0

    def read_records(self) -> collections.deque:
        """
        This method reads the records from the /stream passed
        into the instance.


        Return
        -------
        _dmap_records : collections.Deque
                Deque list of DmapRecords (ordered dictionary)


        See Also
        --------
        DmapScalar : DMap record's scalar data structure
        DmapArray  : DMap record's array data structure

        See DEVELOPER_README.md for more information on
        Dmap data structure.
        """

        # read bytes until end of byte array
        pydarn_log.debug("Reading DMap records")
        while self.cursor < self.dmap_end_bytes:
            new_record = self.read_record()
            self._dmap_records.append(new_record)
            self.rec_num += 1

        self.bytes_check(self.cursor, "cursor",
                         self.dmap_end_bytes, "total bytes in the file")

        self._records = dmap2dict(self._dmap_records)
        return self._records

    def read_record(self) -> collections.OrderedDict:
        """
        Reads a single DMap record from the byte array.

        Return
        ------
        Ordered dict:
                Ordered dictionary containing DmapScalars and DmapArrays from
                a dmap array. The key is the name of the scalar/array and
                value is the DmapScalar/DmapArray data structure

        Raises
        ------
        CursorError
           if the beginning value of the cursor -
           the current position != block size

        See Also
        --------
        read_scalar : reads dmap scalars to be stored in the record
        read_array : reads dmap arrays to be stored in the record
        """
        # used in a cursor check.
        start_cursor_value = self.cursor

        # Need to get encoder for cursor movement though it is not moved.
        # encoding_identifier = self.read_data('i',4)
        # WARNING: normally would call encoding_identifier but since it is not
        # used the extra function call is point less so best just to move the
        # cursor for performance :)
        self.cursor += 4
        block_size = self.read_data('i', 4)

        pydarn_log.debug("Reading Record {record}"
                         .format(record=len(self._dmap_records)))

        # adding 8 bytes because code+size are part of the record.
        # 4 is the number bytes for int format
        remaining_bytes = self.dmap_end_bytes - self.cursor + 2 *\
            DMAP_DATA_TYPES[INT][1]
        self.bytes_check(block_size, "block size",
                         remaining_bytes, "remaining bytes")
        self.zero_negative_check(block_size, "block size")

        num_scalars = self.read_data('i', 4)
        num_arrays = self.read_data('i', 4)

        self.zero_negative_check(num_scalars, "number of scalars")
        self.zero_negative_check(num_arrays, "number of arrays")
        self.bytes_check(num_scalars + num_arrays,
                         "number of scalars + arrays",
                         block_size, "block size")

        pydarn_log.debug("Reading record: reading scalars\n")
        record = collections.OrderedDict()
        for i in range(num_scalars):
            scalar = self.read_scalar()
            record[scalar.name] = scalar

        pydarn_log.debug("Reading record: reading arrays\n")
        for i in range(num_arrays):
            array = self.read_array(block_size)
            record[array.name] = array

        # check for a cursor error
        if (self.cursor - start_cursor_value) != block_size:
            raise dmap_exceptions.CursorError(self.cursor, block_size,
                                              self.rec_num)

        return record

    def read_scalar(self) -> DmapScalar:
        """
        Reads a scalar and stores the properties into a namedtuple DmapScalar.

        Return
        ------
        DmapScalar: namedtuple
            data structure that contains the data properties of
            the scalar read in

        Raises
        ------
        DmapDataError
            if the data type format is DMap
            NOTE: In RST, this is allowed, if an example shows up where this is
            allowed s raise as an issue in the GitHub so the code can
            be re-accessed.

        See Also
        --------
        check_data_type : for other possible raised exceptions
        read_data : reads the data stored in the byte array
        """

        # String and char have a byte size of 1
        scalar_name = self.read_data('s', 1)
        scalar_type = self.read_data('c', 1)

        self.check_data_type(scalar_type, scalar_name)

        scalar_type_fmt = DMAP_DATA_TYPES[scalar_type][0]
        scalar_fmt_byte = DMAP_DATA_TYPES[scalar_type][1]

        if scalar_type_fmt != DMAP:
            scalar_value = self.read_data(scalar_type_fmt, scalar_fmt_byte)
        else:
            message = "Error: Trying to read DMap data type for a scalar."\
                " Failed at record {}".format(self.rec_num)
            # Not sure when this is used in a
            # so better to raise an error if used re-access the code.
            raise dmap_exceptions.DmapDataError(self.dmap_file, message)

        return DmapScalar(scalar_name, scalar_value,
                          scalar_type, scalar_type_fmt)

    def read_array(self, record_size) -> DmapArray:
        """
        Reads an array from a DMap record the byte arrays and
        stores the data properties in a DmapArray structure.

        Return
        ------
        DmapArray : namedtuple
             data structure that contains the data properties of
             the array read in.

        Raises
        -------
            DmapDataError
                if the array properties (like a dimension size == 0)
                are incorrect.

        See Also
        --------
        read_data : reads the data in the byte array
        read_numerical_array : reads in a numerical array from the byte array
        read_string_array: reads in a string/DMap array
        """
        array_name = self.read_data('s', 1)
        array_type = self.read_data('c', 1)

        self.check_data_type(array_type, array_name)

        array_type_fmt = DMAP_DATA_TYPES[array_type][0]
        array_fmt_bytes = DMAP_DATA_TYPES[array_type][1]

        array_dimension = self.read_data('i', 4)
        self.bytes_check(array_dimension, "array dimension",
                         record_size, "record size")
        self.zero_negative_check(array_dimension,
                                 "{name} array dimension"
                                 "".format(name=array_name))

        array_shape = [self.read_data('i', 4)
                       for i in range(0, array_dimension)]
        if array_dimension > 1:
            array_shape.reverse()

        # slist is the array that holds the range gates that have valid data
        # when qflg is 1
        if any(x <= 0 for x in array_shape) and array_name != "slist":
            message = "Error: Array shape {shape} contains "\
                "dimension size <= 0."\
                " Failed at record {rec}".format(shape=array_shape,
                                                 rec=self.rec_num)
            raise dmap_exceptions.DmapDataError(self.dmap_file, message)

        for i in range(array_dimension):
            if array_shape[i] >= record_size:
                message = "Error: Array {index}-dimension size {size}"\
                    " exceeds record size: {rec_size}. "\
                    "Failed at record {rec}"\
                    "".format(index=i,
                              size=array_shape[i],
                              rec_size=record_size,
                              rec=self.rec_num)
                raise dmap_exceptions.DmapDataError(self.dmap_file, message)

        # We could use np.prod(array_shape) but the for loop has a better
        # time performance. Note: cells can also be read as number of elements
        # depending on your background.
        total_num_cells = 1
        for i in array_shape:
            total_num_cells *= i
        self.bytes_check(total_num_cells, "total number of cells",
                         record_size, "record size")

        total_num_cells_bytes = total_num_cells * array_fmt_bytes
        self.bytes_check(total_num_cells_bytes,
                         "total number of cells in bytes",
                         record_size, "record size")

        # parsing an array of strings requires a different method. Numpy can't
        # parse strings or dmaps into arrays the way it can for other
        # types because it doesn't
        # know the sizes. They have to be manually read the slow way.
        # Because chars
        # are encoded as hex literals, they have to be read one at a
        # time to make sense.

        if array_type_fmt == 's':
            message = "Error: Trying to read array of strings."\
                " Currently not implemented."\
                " Failed at record {}".format(self.rec_num)
            # Not sure when this is used in a
            # so better to raise an error if used re-access the code.
            raise dmap_exceptions.DmapDataError(self.dmap_file, message)
            # FIXME: Not working
            # array_value = self.read_string_array(array_shape,
            #                                     array_type_fmt,
            #                                     array_fmt_bytes)
        elif array_type == DMAP:
            message = "Trying to read DMap array data type."\
                " Failed at record {}".format(self.rec_num)
            # Not sure when this is used in a
            # so better to raise an error if used re-access the code.
            raise dmap_exceptions.DmapDataError(self.dmap_file, message)
        else:
            array_value = self.read_numerical_array(array_shape,
                                                    array_type_fmt,
                                                    total_num_cells,
                                                    array_fmt_bytes)

        return DmapArray(array_name, array_value, array_type, array_type_fmt,
                         array_dimension, array_shape)

    def read_data(self, data_type_fmt: str, data_fmt_bytes: int):
        """
        Reads in individual data type from the byte array
        Given a dmap data type.

        Parameter
        ---------
        data_type : int
            dmap data type numerical value for knowing how many bytes to read.

        Return
        ------
        data : data_type
            returns the data that was store in the byte array.
            Has the data type of the dmap data type passed in.

        Raises
        ------
        CursorError
            if cursor is not correctly aligned or at the correct position.
        DmapDataError
            if trying to read a DMap data type

        See Also
        --------
        Struct.unpack : converts the bytes array values from a C structure to
        python objects

        See DEVELOPER_README.md for more information on DMap data types and
        Struct.unpack.
        """
        if self.cursor >= self.dmap_end_bytes:
            message = "Error: Cursor {cursor} extends"\
                " out of buffer {end_bytes}. "\
                "Fails at record {rec}"\
                "".format(cursor=self.cursor,
                          end_bytes=self.dmap_end_bytes,
                          rec=self.rec_num)
            raise dmap_exceptions.CursorError(self.cursor, message=message)

        cursor_offset = self.dmap_end_bytes - self.cursor
        if cursor_offset < data_fmt_bytes:
            message = "Error: Byte offsets {offset} into buffer {fmt}"\
                " are not aligned. Failed at record "\
                "{rec}".format(offset=cursor_offset,
                               fmt=data_fmt_bytes,
                               rec=self.rec_num)
            raise dmap_exceptions.CursorError(self.cursor, message=message)

        # struct.unpack is a python method the converts C struct
        # as python bytes to python values. See DEVELOPER_README.md
        # for more information.
        # TODO: how to handle this gracefully? Return None and check later than
        #       continue in the loop?
        if data_type_fmt == 'c':
            # This casting is needed to keep the correct instance otherwise
            # python will default it to int which later could have some
            # consequences if you try to rewrite the file
            data = self.dmap_bytearr[self.cursor]
            self.cursor += data_fmt_bytes
            return data
        elif data_type_fmt == 's':
            byte_counter = 0
            while self.dmap_bytearr[self.cursor + byte_counter] != 0 or \
                    self.cursor + byte_counter >= self.dmap_end_bytes:
                # add 1 byte to byte_counter because
                # a string is a list of chars
                byte_counter += 1  # 1 is the number of bytes of a char

            self.bytes_check(self.cursor + byte_counter, 'cursor',
                             self.dmap_end_bytes, 'Total number of Bytes')

            # format byte counter to a string format for unpacking
            # WARNING: not using format for performance purposes
            char_count = '%ss' % byte_counter
            data = struct.unpack_from(char_count,
                                      self.dmap_buffer,
                                      self.cursor)
            # 1 is the number of bytes in a char
            self.cursor = self.cursor + byte_counter + 1
            return data[0].decode('utf-8')
        else:
            data = struct.unpack_from(data_type_fmt,
                                      self.dmap_buffer,
                                      self.cursor)
            self.cursor += data_fmt_bytes
            return data[0]

    # FIXME: Currently this method is not working, thus there is no support
    # for array[str], but may not be important. ¯\_(ツ)_/¯
    # The problem is the padding between various string encodings:
    # UTF-8 -> 4 bytes and ASCII -> 1 byte
    # so either DMap would have to incorporate that information plus
    # string length would be a god send... If not one possible solution
    # you would record the length of the record size and subtract the byte
    # (or start at cursor point and increment to the end of the record size)
    # information: encoding, data type, dimension, shape, name, then you
    # should have the length of the array. Then by trial and error (unless we
    # store the encoding information) determine the encoding size, if there is
    # padding you will get "\x00", from there you can jump over the byte size
    # and when a comma occurs then that is the end of your string and append to
    # the array and continue until you reach the end of the record size.
    def read_string_array(self, shape: list,
                          array_type_fmt: str,
                          array_fmt_bytes: int) -> np.ndarray:
        """
        Reads and builds a N-D array of string data types.

        Parameter
        ---------
        shape : list
            list of the arrays dimension size
        array_fmt_bytes : int
            String or char

        Return
        ------
        np.ndarray
            an numpy N-D array of string read in from the byte array
        """

        dim_data = []
        dim = shape
        dimension = dim.pop()
        # FIXME: still need to find a method for testing this... Or a file
        #        that contains a string/DMap array
        # for dim_size in shape:
        #    for i in range(dim_size):
        #        result = self.read_data(array_type_fmt, array_fmt_bytes)
        #        data.append(result)
        # try to avoid using recursion, it is slow
        if not dim:
            dim_data = [self.read_data(array_type_fmt, array_fmt_bytes)
                        for i in range(0, dimension)]
        else:
            dim_data = [self.build_n_dimension_list(list(dim), array_type_fmt)
                        for i in range(0, dimension)]
        return np.array(dim_data)

    def read_numerical_array(self, shape: list, data_type_fmt: str,
                             total_number_cells: int,
                             data_fmt_bytes: int) -> np.ndarray:
        """
        Reads in an array from the byte array into a N-D array.

        Parameter
        ---------
        shape : list
            A list of the array's dimension sizes
        data_type : int
            DMap data type numeric value
        total_number_cells : int
            The number of cells in the N-D array being read in

        Return
        -------
        array : np.ndarray
            The numpy N-D array from the byte array

        Raises
        -------
        DmapDataError
            Mismatch on the data type and the size specified in the byte array.
        """
        if data_type_fmt == 'c':
            # In RST rtypes.h file, chars are defined as int8
            # allowing this assumption to be allowed for now
            array = np.frombuffer(self.dmap_buffer, np.int8,
                                  total_number_cells, self.cursor)
        else:
            array = np.frombuffer(self.dmap_buffer, data_type_fmt,
                                  total_number_cells, self.cursor)
        self.cursor += total_number_cells * data_fmt_bytes
        array = np.reshape(array, shape, order='C')
        return array


class DmapWrite(object):
    """
    Writes DMap records to file or stream and writes SuperDARN file format.
    ...

    Attributes
    -----------
    dmap_records : List[dict]
        List of DMap records
    filename : str
        Name of the file the user wants to write to
    dmap_bytearr : bytearray
        Byte array representing the DMap records in bytes

    Methods
    -------
    write_dmap(filename)
        Writes DMap records to DMap format with the given filename
    write_dmap_stream(dmap_records)
        Writes DMap records to DMap format byte stream
    dict_key_diff(dict1, dict2)
        Returns a set of the difference between dict1 and dict2 keys
    missing_field_check(file_struct_list, record, rec_num)
        Checks if there is any missing fields in the record from
        a list of possible file fields
    extra_field_check(file_struct_list, record, rec_num)
        Checks if there is any extra fields in the record from
        a list of possible file fields
    incorrect_types_check(file_struct_list, record)
        Checks if there is any incorrect types in the record from
        a list of possible file
        fields and their data type formats
    dict_list2set(dict_list)
        Converts a list of dictionaries to a set containing their keys
    SuperDARN_file_structure_to_bytes(file_struct_list)
        Converts DMap records to SuperDARN file structure bytes based
        on file_struct_list
    dmap_records_to_bytes()
        Converts DMap records to byte array stored in dmap_bytearr
    dmap_scalar_to_bytes(scalar)
        Converts a DmapScalar to bytes
    dmap_array_to_bytes(array)
        Converts a DmapArray to bytes
    """
    def __init__(self, dmap_records: List[dict] = [], filename: str = ""):
        """
        Writes DMap records to a given filename of byte array in DMap format,
        this includes the following SuperDARN file types:
                                                      - Iqdat
                                                      - Rawacf
                                                      - Fitacf
                                                      - Grid
                                                      - Map

        Parameters
        ----------
        dmap_records : List[dict]
            list of dictionaries representing a list of DMap records containing
            DmapScalar and DmapArrays
        filename : str
            The path and name of the file the user wants to write to
        dmap_file_fmt : str
             types, the following are supported:
                                     - 'iqdat' : SuperDARN file type
                                     - 'rawacf' : SuperDARN file type
                                     - 'fitacf' : SuperDARN file type
                                     - 'grid' : SuperDARN file type
                                     - 'map' : SuperDARN file type
                                     - 'dmap' : writes a file in DMap format
                                     - 'stream' : writing to dmap data stream

        Raises
        ------
        DmapTypeError
        FilenameRequiredError
        """
        self.filename = filename
        self.rec_num = 0

        self.dmap_records = self.__check_dmap_dict(dmap_records)

        self.dmap_bytearr = bytearray()
        pydarn_log.debug("Initiating DmapWrite")

    def __repr__(self):
        """ for representation of the class object"""
        # __class__.__name__ allows to grab the class name such that
        # when a class inherits this one, the class name will be the child
        # class and not the parent class (dmap classes)
        return "{class_name}({filename}, {rec_num})"\
               "".format(class_name=self.__class__.__name__,
                         filename=self.filename,
                         rec_num=self.rec_num)

    def __str__(self):
        """ for printing of the class object"""
        # __class__.__name__ allows to grab the class name such that
        # when a class inherits this one, the class name will be the child
        # class and not the parent class (dmap classes)
        return "Writing to filename: {filename} at record number: {rec_num}"\
               "".format(filename=self.filename,
                         rec_num=self.rec_num)

    def __check_dmap_dict(self, dmap_records):
        """
        Checks if the dmap data is a dict, if so, converts to a dmap_structure

        Parameter:
            dmap_data : List[dict]
                dmap_data structure to test if it is a dict of value or dmap
                structure
        Returns:
            dmap_structure : List[dict]
                returns the dmap structure of dmap_data
        """
        try:
            first_value = list(dmap_records[0].values())[0]
            if isinstance(first_value, DmapScalar) or isinstance(first_value,
                                                                 DmapArray):
                return dmap_records
            else:
                return dict2dmap(dmap_records)
        except IndexError:
            return dmap_records

    # HONEY BADGER method: Because dmap just don't care -
    # will write anything into DMap format with out any file
    # structure checks. See supedarn.py for methods that do checks
    # for superDARN file structures.
    def write_dmap(self, filename: str = ""):
        """
        Writes DMap record to  format.

        Parameters:
        -----------
        filename : str
            The name of the DMap file including the path

        WARNING:
        --------
        No checks are done, this up to the user to ensure their fields are
        correct.
        """
        self._filename_check(filename)
        pydarn_log.debug("Writing : {}".format(self.filename))
        self.write_dmap_stream()
        with open(self.filename, 'wb') as f:
            f.write(self.dmap_bytearr)

    def write_dmap_stream(self, dmap_records: List[dict] = []) -> bytearray:
        """
        Writes DMap record to  format.

        Return
        ------
        dmap_bytearr : bytearray
            Bytearray of the DMap records

        WARNING:
        --------
        No checks are done, this up to the user to ensure their fields are
        correct.
        """
        if self.dmap_records == []:
            self.dmap_records = self.__check_dmap_dict(dmap_records)
        self._empty_record_check()
        pydarn_log.debug("Writing to dmap stream")
        self.dmap_records_to_bytes()
        return self.dmap_bytearr

    def _empty_record_check(self):
        if self.dmap_records == []:
            raise dmap_exceptions.DmapDataError(self.filename,
                                                "DMap record is empty "
                                                "there is nothing to write.")

    def _filename_check(self, filename: str = ""):
        """
        Checks if a filename is given and overrides the current
        filename with the new one.

        Parameter:
        ----------
        filename : str
            Name of the file to write to

        Raises:
        ------
        FilenameRequireError - raised if no filename is given
        """
        if self.filename == "" and filename == "":
            raise dmap_exceptions.FilenameRequiredError()
        elif filename != "":
            self.filename = filename

    def dmap_records_to_bytes(self):
        """
        Loops through the DMap records and calls dmap_record_to_bytes
        to convert the DMap records to a byte array.

        Future use of this function is for parallelization.
        """
        # For performance increase len of the records can be
        # attribute value initialized in the class
        pydarn_log.debug("Converting DMap records to bytes")
        for self.rec_num in range(len(self.dmap_records)):
            record = self.dmap_records[self.rec_num]
            self._dmap_record_to_bytes(record)

    def _dmap_record_to_bytes(self, record: dict):
        """
        Converts DMap record to byte stream and stores in the
        dmap byte array

        Parameter
        ---------
        record : dict
            DMap record

        Notes
        -----
        Might be useful to return bytes of records for potential
        writing into a data stream or real-time stream
        """
        # DaVitpy reference, still unclear of meaning of the numeber
        encoding_identifier = 65537
        num_scalars = 0
        num_arrays = 0
        data_bytearray = bytearray()
        for name, data_info in record.items():
            if isinstance(data_info, DmapScalar):
                data_bytearray.extend(self.dmap_scalar_to_bytes(data_info))
                num_scalars += 1
            elif isinstance(data_info, DmapArray):
                data_bytearray.extend(self.dmap_array_to_bytes(data_info))
                num_arrays += 1
            else:
                raise dmap_exceptions.DmapTypeError(self.filename,
                                                    type(data_info),
                                                    self.rec_num)

        # 16 = encoding_identifier (int - 4 bytes) + num_scalars (int - 4) +
        #      num_arrays (int - 4) + size of the block (int - 4)
        block_size = len(data_bytearray) + 16

        # TODO: potential performance increase by saving the extend object
        self.dmap_bytearr.extend(struct.pack('i', encoding_identifier))
        self.dmap_bytearr.extend(struct.pack('i', block_size))
        self.dmap_bytearr.extend(struct.pack('i', num_scalars))
        self.dmap_bytearr.extend(struct.pack('i', num_arrays))
        self.dmap_bytearr.extend(data_bytearray)

    def dmap_scalar_to_bytes(self, scalar: DmapScalar) -> bytes:
        """
        Converts a DmapScalar to byte format.
        Byte format: name, data type, data

        Parameter
        ---------
        scalar : DmapScalar
            Dmap scalar to be convert into byte format

        Return
        ------
        scalar_total_bytes : bytes
            Bytes of the scalar
        """

        scalar_name = "{0}\0".format(scalar.name)
        scalar_name_format = '{0}s'.format(len(scalar_name))
        scalar_name_bytes = struct.pack(scalar_name_format,
                                        scalar_name.encode('utf-8'))

        scalar_type_bytes = struct.pack('c',
                                        chr(scalar.data_type).encode('utf-8'))

        if scalar.data_type_fmt == 's':
            scalar_data = "{0}\0".format(scalar.value)
            scalar_data_format = '{0}s'.format(len(scalar_data))
            scalar_data_bytes = struct.pack(scalar_data_format,
                                            scalar_data.encode('utf-8'))
        elif scalar.data_type_fmt == 'c':
            # char is defined as int8 by RST
            if isinstance(scalar.value, str):
                raise dmap_exceptions.DmapCharError(scalar.name, self.rec_num)
            # scalar_data_bytes = scalar.value
            scalar_data_bytes = chr(scalar.value).encode('utf-8')
        elif scalar.data_type_fmt == 'f':
            scalar_data_value = np.float32(scalar.value)
            scalar_data_bytes = struct.pack(scalar.data_type_fmt,
                                            scalar_data_value)
        else:
            scalar_data_bytes = struct.pack(scalar.data_type_fmt, scalar.value)

        scalar_total_bytes = scalar_name_bytes + scalar_type_bytes +\
            scalar_data_bytes

        return scalar_total_bytes

    def dmap_array_to_bytes(self, array: DmapArray) -> bytes:
        """
        Converts a DmapArray to the byte format.

        Byte format: name, data type, dimension, shape, data

        Parameter
        ---------
        array : DmapArray
            Dmap array to be converted to bytes

        Return
        -------
        array_total_bytes : bytes
            Total bytes of the array
        """
        # need to ensure the array is flattened
        if array.data_type_fmt == 'c' and isinstance(array.value[0], str):
            raise dmap_exceptions.DmapCharError(array.name, self.rec_num)
        elif array.data_type_fmt == 's':
            message = "Error: Trying to read array of strings."\
                " Currently not implemented."\
                " Failed at record {}".format(self.rec_num)
            raise dmap_exceptions.DmapDataError(self.filename, message)

        array_value = array.value.flatten()
        array_name = "{0}\0".format(array.name)
        array_name_format = '{0}s'.format(len(array_name))
        array_name_bytes = struct.pack(array_name_format,
                                       array_name.encode('utf-8'))

        array_type_bytes = struct.pack('c',
                                       chr(array.data_type).encode('utf-8'))

        array_dim_bytes = struct.pack('i', array.dimension)
        array_shape_bytes = bytes()
        if array.dimension > 1:
            array.shape.reverse()
        for size in array.shape:
            array_shape_bytes += struct.pack('i', size)

        array_data_bytes = array_value.tostring()

        array_total_bytes = array_name_bytes + array_type_bytes + \
            array_dim_bytes + array_shape_bytes + array_data_bytes

        return array_total_bytes



# Copyright (C) SuperDARN Canada, University of Saskatchewan
# Authors: Keith Kotyk and Marina Schmidt
import logging
pydarn_logger = logging.getLogger('pydarn')


class SuperDARNFileTypeError(Exception):
    """
    SuperDARN file type that is not implement or incorrect

    Parameter
    --------
    filename : str
        name of the file associated to the wrong SuperDARN file type
    file_type : str
        SuperDARN file type that is not implement or correct
    """
    def __init__(self, filename: str, file_type: str):
        self.filename = filename
        self.file_type = file_type
        self.message = "Error: {file_type} is not a SuperDARN file "\
            "format type."\
            "{filename} was not created. Please check the spelling of"\
            " the file type is correct or is"\
            " implemented.".format(file_type=self.file_type,
                                   filename=self.filename)
        super().__init__(self.message)
        pydarn_logger.error(self.message)


class SuperDARNFieldMissingError(Exception):
    """
    Raised when a field is missing from data that is required of the SuperDARN
    file type fields

    Parameter
    ---------
    record_num : int
        The record number associated to the error
    fields : set
        set of fields that are missing from the record
    """
    def __init__(self, record_num: int, fields: set):
        self.record_number = record_num
        self.fields = fields
        self.message = "Error: The following fields in record {num} "\
            "are missing:"\
            " {fields}".format(num=self.record_number,
                               fields=self.fields)
        super().__init__(self.message)
        pydarn_logger.error(self.message)


class SuperDARNExtraFieldError(Exception):
    """
    Raised when a extra field is in the data that is not in the SuperDARN file
    type fields

    Parameter
    ---------
    record_num : int
        record number associated to the error
    fields : set
        set of the fields that are extra in the record
    """
    def __init__(self, record_num, fields):
        self.record_number = record_num
        self.fields = fields
        self.message = "Error: The following fields in record {num} are"\
            " not allowed:"\
            " {fields}".format(num=self.record_number,
                               fields=self.fields)
        super().__init__(self.message)
        pydarn_logger.error(self.message)


class SuperDARNDataFormatTypeError(Exception):
    """
    Raised when a data format type does not match the SuperDARN file type
    data types

    Parameter
    --------
    incorrect_type : set
        set of the fields that have incorrect types
    record_num : int
        record number of the associated to the error
    """
    def __init__(self, incorrect_types: set, record_num: int):
        self.incorrect_params = incorrect_types
        self.record_number = record_num
        self.message = "Error: In record {num}, following parameters"\
            " need to be the data type:"\
            " {incorrect}".format(num=self.record_number,
                                  incorrect=self.incorrect_params)
        super().__init__(self.message)
        pydarn_logger.error(self.message)



# Copyright 2019 SuperDARN Canada, University of Saskatchewan
# Author: Marci Detwiller
"""
This file contains a utility class for Borealis file type
checks.

Classes
-------
BorealisUtilities: utilites class that contains static methods for
SuperDARN Borealis file type checking

Exceptions
----------
BorealisFieldMissingError
BorealisExtraFieldError
BorealisDataFormatTypeError
BorealisNumberOfRecordsError

See Also
--------
For more information on Borealis data files and how they convert to dmap,
see: https://borealis.readthedocs.io/en/latest/

Future Work
-----------
Look into combining SDARNUtilties and BorealisUtilities into a single util
class

"""
import logging
import numpy as np
import sys

from functools import reduce
from typing import Union, List

from pydarn import borealis_exceptions

pydarn_log = logging.getLogger('pydarn')


class BorealisUtilities():
    """
    Utility class containing static methods used by other classes.

    Contains static methods that do dictionary set calculations
    used for determining if there is any missing or extra
    fields in Borealis file types. Also does data format type checks
    for Borealis file types.

    Static Methods
    --------------
    set_diff(dict1, dict2)
        Returns a set of the difference between dict1 and dict2 keys or the
        elements of the dict1 and dict2 sets/lists
    dict_list2set(dict_list)
        Converts a list of dictionaries to a set containing their keys
    missing_field_check(filename, file_struct_list, parameter_names,
            [record_name])
        Checks if there is any missing fields in the record/file from
        a list of possible file fields
    extra_field_check(filename, file_struct_list, parameter_names,
            [record_name])
        Checks if there is any extra fields in the record/file from
        a list of possible file fields
    record_incorrect_types_check(filename, attributes_type_dict,
            datasets_type_dict, record, record_name)
        Checks if there is any incorrect types in the record from
        a list of possible file fields and their data type formats
    array_incorrect_types_check(filename, attributes_type_dict,
            datasets_type_dict, unshared_parameters, file_data)
        Checks if there are any incorrect types in the file data
        from a list of possible fields and their formats, and whether
        the data is a shared field in the array file or not.
    array_num_records_check(filename, unshared_parameters, file_data)
        Checks if there is a consistent number of records in the arrays
        from the array file.
    """

    @staticmethod
    def set_diff(dict1: Union[dict, set, list],
                 dict2: Union[dict, set, list]) -> set:
        """
        Determines the difference in the key set from the
        first dictionary to the second dictionary.
        ex) Let A = {a, b, c} and B = {d, a, b}
        Then A - B = {c}

        Parameters
        ----------
        dict1: dict or set
            dictionary or set to subtract from
        dict2: dict or set
            dictionary or set subtracting from dict1

        Returns
        -------
        dict_keys: set
            difference between dict1 and dict2 keys or the sets/lists
        """
        diff_keys = set(dict1) - set(dict2)
        return diff_keys

    @staticmethod
    def dict_list2set(dict_list: List[dict]) -> set:
        """
        Converts a list of dictionaries to list of sets

        Parameters
        ----------
        dict_list: list
            list of dictionaries

        Returns
        -------
        complete_set: set
            set containing all dictionary key from the list of dicts
        """
        # convert dictionaries to set to do some set magic
        sets = [set(dic) for dic in dict_list]
        # create a complete set list
        # * - expands the list out into multiple set arguments
        # then the union operator creates it into a full unique set
        # example:
        # s = [{'a','v'}, {'v','x'}] => set.union(*s) = {'a', 'v', 'x'}
        complete_set = set.union(*sets)
        return complete_set

    @staticmethod
    def missing_field_check(filename: str, file_struct_list: List[dict],
                            parameter_names: Union[List[str], dict, set],
                            **kwargs):
        """
        Checks if any fields are missing from the file/record compared to the
        file structure.

        Parameters
        ----------
        filename: str
            Name of the file being checked
        file_struct_list: List[dict]
            List of dictionaries for the possible file structure fields
        parameter_names: List[str], dict, set
            List of parameter names, or set, or dict, in the file or in
            the record.
        record_name: str
            The name of the record (first sequence start time), if in a
            record style file.

        Raises
        -------
        BorealisFieldMissingError
        BorealisStructureError

        Notes
        -----
        Checks sets and subsets. Any missing fields are a problem because
        Borealis field names are well-defined.
        """
        missing_fields = set()
        for file_struct in file_struct_list:
            diff_fields = BorealisUtilities.set_diff(file_struct,
                                                     parameter_names)
            if len(diff_fields) != 0:
                missing_fields = missing_fields.union(diff_fields)

        if len(missing_fields) == \
                reduce(lambda a, b: a+b, [len(x) for x in file_struct_list]):
            # all fields are missing
            if 'record_name' in kwargs.keys():
                raise borealis_exceptions.\
                        BorealisStructureError("All fields expected are "
                                               "missing in record {}: {}"
                                               "".format(kwargs['record_name'],
                                                         missing_fields))
            else:
                raise borealis_exceptions.\
                        BorealisStructureError("All fields expected are"
                                               " missing : {}"
                                               "".format(missing_fields))

        if len(missing_fields) > 0:
            if 'record_name' in kwargs.keys():
                raise borealis_exceptions.\
                        BorealisFieldMissingError(filename, missing_fields,
                                                  record_name=kwargs['record_name'])
            else:
                raise borealis_exceptions.\
                        BorealisFieldMissingError(filename,
                                                  missing_fields)

    array_missing_field_check = record_missing_field_check =\
        missing_field_check

    @staticmethod
    def extra_field_check(filename: str, file_struct_list: List[dict],
                          parameter_names: Union[List[str], dict, set],
                          **kwargs):
        """
        Check if there is an extra field in the file/record.

        Parameters
        ----------
        filename: str
            Name of the file being checked
        file_struct_list: List[dict]
            List of dictionaries for the possible file structure fields
        parameter_names: List[str], dict, set
            List of parameter names, or set, or dict, in the file or in
            the record.
        record_name: str
            Record name for better error message information, if in a
            record style file.

        Raises
        ------
        BorealisExtraFieldError
        """
        file_struct = BorealisUtilities.dict_list2set(file_struct_list)
        extra_fields = BorealisUtilities.set_diff(parameter_names, file_struct)

        if len(extra_fields) > 0:
            if 'record_name' in kwargs.keys():
                raise borealis_exceptions.\
                        BorealisExtraFieldError(filename,
                                                extra_fields,
                                                record_name=kwargs['record_name'])
            else:
                raise borealis_exceptions.\
                        BorealisExtraFieldError(filename, extra_fields)

    array_extra_field_check = record_extra_field_check = extra_field_check

    @staticmethod
    def record_incorrect_types_check(filename: str, attributes_type_dict: dict,
                                     datasets_type_dict: dict, record: dict,
                                     record_name: str):
        """
        Checks if the record's data type formats are correct according
        to the file structure fields data type formats.

        Checks both single element types and numpy array dtypes separately.

        Parameters
        ----------
        filename: str
            Name of the file being checked
        attributes_type_dict: dict
            Dictionary with the required types for the attributes in the file.
        datasets_type_dict: dict
            Dictionary with the require dtypes for the numpy array types in the
            file.
        record: dict
            Record dictionary
        record_name: str
            Record name for a better error message information

        Raises
        ------
        BorealisDataFormatTypeError
        """
        incorrect_types_check = {param: str(attributes_type_dict[param])
                                 for param in attributes_type_dict.keys()
                                 if type(record[param]) !=
                                 attributes_type_dict[param]}

        incorrect_types_check.update({param: 'np.ndarray of ' +
                                      str(datasets_type_dict[param])
                                      for param in datasets_type_dict.keys()
                                      if record[param].dtype.type !=
                                      datasets_type_dict[param]})
        if len(incorrect_types_check) > 0:
            raise borealis_exceptions.BorealisDataFormatTypeError(
                filename, incorrect_types_check, record_name=record_name)

    @staticmethod
    def array_incorrect_types_check(filename: str, attributes_type_dict: dict,
                                    datasets_type_dict: dict, file_data: dict):
        """
        Checks if the file's data type formats are correct according
        to the file structure fields data type formats and whether or
        not the field is a shared field between records.

        Checks both single element types and numpy array dtypes separately.

        Parameters
        ----------
        filename: str
            Name of the file being checked
        attributes_type_dict: dict
            Dictionary with the required types for the attributes in the file.
        datasets_type_dict: dict
            Dictionary with the require dtypes for the numpy
            arrays in the file.
        file_data: dict
            dictionary containing all file information.

        Raises
        ------
        BorealisDataFormatTypeError
        BorealisStructureError
        """

        attributes_type_dict_keys = sorted(list(attributes_type_dict.keys()))
        np_array_types = [isinstance(file_data[param], np.ndarray) for param
                          in attributes_type_dict_keys]

        # There is a better way to do this TODO
        array_keys = [attributes_type_dict_keys[x] for x, y in
                      enumerate(np_array_types) if y]  # True
        array_fields = [(field, attributes_type_dict[field]) for field in
                        array_keys]

        if any(np_array_types):
            raise borealis_exceptions.\
                    BorealisStructureError('Unexpected array in the '
                                           'following fields where single'
                                           ' attributes expected:  {}'
                                           ''.format(array_fields))

        incorrect_types_check = {param: str(attributes_type_dict[param])
                                 for param in attributes_type_dict.keys()
                                 if type(file_data[param]) !=
                                 attributes_type_dict[param]}

        datasets_type_dict_keys = sorted(list(datasets_type_dict.keys()))
        np_array_types = [isinstance(file_data[param], np.ndarray) for param in
                          datasets_type_dict_keys]
        array_keys = [datasets_type_dict_keys[x] for x, y in
                      enumerate(np_array_types) if not y]
        non_array_fields = [(field, datasets_type_dict[field]) for field in
                            array_keys]

        if not all(np_array_types):
            raise borealis_exceptions.\
                    BorealisStructureError('Unexpected single attribute in the'
                                           ' following fields where arrays of '
                                           'given dtype are expected: {}'
                                           ''.format(non_array_fields))

        incorrect_types_check.update({param: 'np.ndarray of ' +
                                      str(datasets_type_dict[param])
                                      for param in datasets_type_dict.keys()
                                      if file_data[param].dtype.type !=
                                      datasets_type_dict[param]})
        if len(incorrect_types_check) > 0:
            raise borealis_exceptions.\
                    BorealisDataFormatTypeError(filename,
                                                incorrect_types_check)

    @staticmethod
    def array_num_records_check(filename: str, unshared_parameters: List[str],
                                file_data: dict):
        """
        Checks the number of records in all unshared parameter fields to
        ensure they are the same for data integrity.

        Parameters
        ----------
        filename: str
            Name of the file being checked
        unshared_parameters: List[str]
            List of parameter names that are not shared between all the records
            in the array restructured file, i.e. will have a dimension = to
            number of records.
        file_data: dict
            dictionary containing all file information.

        Raises
        ------
        BorealisNumberOfRecordsError
        """
        num_records = {parameter: file_data[parameter].shape for parameter
                       in unshared_parameters}

        dimensions = list(num_records.values())
        try:
            if not all(x[0] == dimensions[0][0] for x in dimensions):
                raise borealis_exceptions.\
                        BorealisNumberOfRecordsError(filename, num_records)
        except IndexError:  # some fields are not arrays!
            tb = sys.exc_info()[2]
            raise borealis_exceptions.\
                BorealisNumberOfRecordsError(filename,
                                             num_records).with_traceback(tb)

    @classmethod
    def check_arrays(cls, origin_string: str, arrays: dict,
                     attribute_types: dict, dataset_types: dict,
                     unshared_fields: List[str]):
        """
        Parameters
        ----------
        origin_string: str
            Name of file to be checked or other origin descriptor.
        arrays: dict
            Dictionary of arrays to be checked.
        attribute_types: dict
            Dictionary with the required types for the attributes in the file.
        dataset_types: dict
            Dictionary with the require dtypes for the numpy arrays in the
            file.
        unshared_fields: List[str]
            List of fields that are not shared between the records and
            therefore should be an array with first dimension = number of
            records

        Raises
        ------
        BorealisFieldMissingError - when a field is missing from the Borealis
                                file
        BorealisExtraFieldError - when an extra field is present in the
                                Borealis file
        BorealisDataFormatTypeError - when a field has the incorrect
                                field type for the Borealis file
        BorealisNumberOfRecordsError - when the number of records cannot
                                be discerned from the arrays

        See Also
        --------
        array_missing_field_check(filename, format_fields, parameters)
                        - checks for missing fields. See this
                        method for information on why we use format_fields.
        array_extra_field_check(filename, format_fields, parameters)
                        - checks for extra fields in the record
        array_incorrect_types_check(filename, attribute_types_dict,
                        dataset_types_dict, unshared_fields, file_data)
                         - checks for incorrect data types for file fields
        array_num_records_check(filename, unshared_fields, file_data)
        """
        all_format_fields = [attribute_types, dataset_types]
        cls.array_missing_field_check(origin_string, all_format_fields, arrays)
        cls.array_extra_field_check(origin_string, all_format_fields, arrays)
        cls.array_incorrect_types_check(origin_string, attribute_types,
                                        dataset_types, arrays)
        cls.array_num_records_check(origin_string, unshared_fields, arrays)

    @classmethod
    def check_records(cls, origin_string: str, records: dict,
                      attribute_types: dict, dataset_types: dict):
        """
        Do checks on the restructured data the same as would be done as if
        the dictionary had just been read from file.

        Parameters
        ----------
        origin_string: str
            Name of file to be checked or other origin descriptor.
        records: dict
            Dictionary of records to be checked for errors.
        attribute_types: dict
            Dictionary with the required types for the attributes in the file.
        dataset_types: dict
            Dictionary with the require dtypes for the numpy arrays in the
            file.

        Raises
        ------
        OSError: file does not exist
        BorealisFieldMissingError - when a field is missing from the Borealis
                                file/stream type
        BorealisExtraFieldError - when an extra field is present in the
                                Borealis file/stream type
        BorealisDataFormatTypeError - when a field has the incorrect
                                field type for the Borealis file/stream type

        See Also
        --------
        record_missing_field_check(filename, format_fields, record,
                        record_name) - checks for missing fields. See this
                        method for information on why we use format_fields.
        record_extra_field_check(filename, format_fields, record, record_name)
                        - checks for extra fields in the record
        record_incorrect_types_check(filename, attribute_types_dict,
                        dataset_types_dict, record, record_name) - checks
                        for incorrect data types for file fields
        """
        all_format_fields = [attribute_types, dataset_types]

        for record_name, record in records.items():
            cls.record_missing_field_check(origin_string, all_format_fields,
                                           record, record_name=record_name)
            cls.record_extra_field_check(origin_string, all_format_fields,
                                         record, record_name=record_name)
            cls.record_incorrect_types_check(origin_string, attribute_types,
                                             dataset_types, record,
                                             record_name)



# Copyright (C) 2020 SuperDARN Canada, University of Saskatchewan
# Authors: Marina Schmidt and Danno Peters

"""
This module contains SuperDARN radar information
"""
import os

from typing import NamedTuple
from enum import Enum
from datetime import datetime, timedelta
from pydarn import radar_exceptions


def read_hdw_file(abbrv, date: datetime = None):
    """
    Reads the hardware file for the associated abbreviation of the radar name.

    Parameters
    ----------
        abbrv : str
            Radars 3 letter assigned abbreviation
        date: datetime
            Datetime object of hardware information to obtain
            default: current date

    Return
    ------
    _HdwInfo object that contains all the field names in a hardware
        file

    Raises
    ------
    HardwareFileNotFoundError raised when there is no hardware file found for
    the given abbreviation
    """
    if date is None:
        date = datetime.now()

    hdw_path = os.path.dirname(__file__)+'/hdw/'
    hdw_file = "{path}/hdw.dat.{radar}".format(path=hdw_path, radar=abbrv)
    try:
        with open(hdw_file, 'r') as reader:
            for line in reader.readlines():
                if '#' not in line and len(line.split()) > 1:
                    hdw_data = line.split()
                    """
                    Hardware files give the year and seconds from the beginning
                    of that year. Thus to check the date if it corresponds we
                    need to convert to a datetime object and then compare.
                    """
                    hdw_line_date = datetime(year=int(hdw_data[1]), month=1,
                                             day=1) +\
                        timedelta(seconds=int(hdw_data[2]))
                    if hdw_line_date > date:
                        """
                        Hardware data array positions definitions:
                            0: station id - stid
                            1: last year that the parameter string is valid.
                            Note: currently updated line will
                            have a year of 2999
                            meaning it is currently still up to date.
                            2: last second of year that parameter
                            string is valid.
                            3: Geographic latitude of radar site
                            4: Geographic longitude of radar site
                            Note: southern lat and long are negative
                            5: Altitude of the radar site (meters)
                            6: Scanning boresight - direction of
                            the centre beam,
                            measured in degrees relative to geographic north.
                            Counter clockwise rotations are negative.
                            7: Beam separation (Angular seperation in degrees)
                            8: velocity sign - at radar level,
                            backscattered signal with
                            frequencies above the transmitted frequency
                            are assigned positive
                            Doppler velocities while backscattered signals
                            with frequencies
                            below the transmitted frequency are assigned
                            negative Doppler
                            velocity. Can be changed in receiver design.
                            9: Analog Rx attenuator step (dB)
                            10: Tdiff - propagation time from
                            interferometer array antenna
                            to phasing matrix input miunus propagation
                            time from main array
                            antenna through transmitter to phasing
                            matrix input.
                            (microseconds)
                            11: phase sign - to account for any cable errors
                            Interferometer offset - displacement of midpoint
                            interferometer array from midpoint main
                            array (meters).
                            12: x direction - along the line of antennas
                            with +X toward
                            higher antenna number
                            13: y direction - along the array normal
                            with +Y in the
                            direction of the array normal
                            14: z direction - is the altitude difference, +Z up
                            15: Analog Rx rise time (microseconds)
                            16: Analog Attenuation stages - gain control of
                            an analog
                            receiver or front-end
                            17: maximum range gates
                            18: maximum number of beams
                        """
                        return _HdwInfo(int(hdw_data[0]), abbrv,
                                        _Coord(float(hdw_data[3]),
                                               float(hdw_data[4]),
                                               float(hdw_data[5])),
                                        float(hdw_data[6]), float(hdw_data[7]),
                                        float(hdw_data[8]), float(hdw_data[9]),
                                        float(hdw_data[10]),
                                        float(hdw_data[11]),
                                        _InterferometerOffset(float(hdw_data[12]),
                                                              float(hdw_data[13]),
                                                              float(hdw_data[14])),
                                        float(hdw_data[15]),
                                        float(hdw_data[16]),
                                        int(hdw_data[17]), int(hdw_data[18]))
    except FileNotFoundError:
        raise radar_exceptions.HardwareFileNotFoundError(abbrv)


class Hemisphere(Enum):
    """
    Class used to denote which hemisphere a radar is located in

    Attributes
    ----------
    Hemisphere.North = 1
    Hemisphere.South = -1

    Notes
    -----
    This is based on the values assigned to the MAP file hemisphere
    field in RST>
    """
    North = 1
    South = -1


class _InterferometerOffset(NamedTuple):
    """
    Named tuple class to contain the interferometer offset
    Cartesian coordinates.

    Attributes
    ----------
    x : float
        direction along the line of antennas with +X toward
        higher antenna number
    y : float
        direction along the array normal with +Y in the
        direction of the array normal
    z : float
        direction is the altitude difference, +Z up
    """
    x: float
    y: float
    z: float


class _Coord(NamedTuple):
    """
    Class used to store Geo Coordinates (utilized for both Geographic
    and Geomagnetic) representing location and Boresite direction

    Attributes
    ----------
    lat : float
        Latitude in decimal format
    lon : float
        Longitude in decimal format
    alt : float
        Altitude in meters
    """
    lat: float
    lon: float
    alt: float


class _HdwInfo(NamedTuple):
    """
    Class used to store relevant information about the SuperDARN Radars

    Attributes
    ----------
    stid : int
        station number
    abbrev : str
        three letter station abbreviation
    geographic : _Coord object
        Named Tuple containing geographic latitude longitude and altitude
    boresight : float
        boresight center beam in degrees
    beam_seperation : float
        angular separation between radar beams
    velocity_sign : float
        at radar level, backscattered signal with
        frequencies above the transmitted frequency are assigned positive
        Doppler velocities while backscattered signals with frequencies
        below the transmitted frequency are assigned negative Doppler
        velocity. Can be changed in receiver design.
    rx_attenuator : float
        Analog Rx attenuator step (dB)
    tdiff : float
        propagation time from interferometer array antenna
        to phasing matrix input minus propagation time from main array
        antenna through transmitter to phasing matrix input.
        (microseconds)
    phase_sign : float
        To account for flipped cable errors
    interferometer_offset : _InterferometerOffset
        displacement of midpoint
        interferometer array from midpoint main array in
        Cartesian coordinates(meters).
    rx_rise : float
        Analog Rx rise time (microseconds)
    attenuation_stages : float
        Analog Attenuation stages - gain control of an analog
        receiver or front-end
    gates :int
        number of range gates per beam
    beams : int
        number of possible beams

    See Also
    --------
    read_hdw_file : function for reading hardware files
    _Coord : object contain coordinate information
    """
    stid: int
    abbrev: str
    geographic: _Coord
    boresight: float
    beam_seperation: float
    velocity_sign: float
    rx_attenuator: float
    tdiff: float
    phase_sign: float
    interferometer_offset: _InterferometerOffset
    rx_rise_time: float
    attenuation_stages: float
    gates: int
    beams: int


class _Radar(NamedTuple):
    """
    Class used to combine all Radar information
    Attributes
    ----------
    name: str
        Name of the Radar
    institute: str
        Institution of the Radar
    hemisphere: Hemisphere
        Hemisphere the radar belongs to
    hardware_info: _HdwInfo
        NamedTuple containing all hardware information from hardware files
    """
    name: str
    institution: str
    hemisphere: Hemisphere
    hardware_info: _HdwInfo


class SuperDARNRadars():
    """
    Class containing a dictionary of Nested Named Tuples with information
    about each radar in SuperDARN.

    Attributes
    ----------
        radars: dict
            dictionary of each SuperDARN radar with key being STID value and
            a _Radar object containing the name, institutional and hardware
            information of the radar

    See Also
    --------
        _Radar : radar object containing radar information
        read_hdw_file : function to read hardware information for a given radar
    """
    # Information obtained from
    # http://vt.superdarn.org/tiki-index.php?page=Radar+Overview
    radars = {209: _Radar('Adak Island East',
                          'University of Alaska Fairbanks', Hemisphere.North,
                          read_hdw_file('ade')),
              208: _Radar('Adak Island West', 'University of Alaska Fairbanks',
                          Hemisphere.North, read_hdw_file('adw')),
              33: _Radar('Blackstone', 'Virginia Tech', Hemisphere.North,
                         read_hdw_file('bks')),
              207: _Radar('Christmas Valley East', 'Dartmouth College',
                          Hemisphere.North, read_hdw_file('cve')),
              206: _Radar('Christmas Valley West', 'Dartmouth College',
                          Hemisphere.North, read_hdw_file('cvw')),
              66: _Radar('Clyde River', 'University of Saskatchewan',
                         Hemisphere.North, read_hdw_file('cly')),
              205: _Radar('Fort Hays East', 'Virginia Tech', Hemisphere.North,
                          read_hdw_file('fhe')),
              204: _Radar('Fort Hays West', 'Virginia Tech', Hemisphere.North,
                          read_hdw_file('fhw')),
              1: _Radar('Goose Bay', 'Virginia Tech', Hemisphere.North,
                        read_hdw_file('gbr')),
              10: _Radar('Hankasalmi', 'University of Leicester',
                         Hemisphere.North, read_hdw_file('han')),
              40: _Radar('Hokkaido East', 'Nagoya University',
                         Hemisphere.North, read_hdw_file('hok')),
              41: _Radar('Hokkaido West', 'Nagoya University',
                         Hemisphere.North, read_hdw_file('hkw')),
              64: _Radar('Inuvik', 'University of Saskatchewan',
                         Hemisphere.North, read_hdw_file('inv')),
              3: _Radar('Kapuskasing', 'Virginia Tech', Hemisphere.North,
                        read_hdw_file('kap')),
              16: _Radar('King Salmon',
                         'National Institute of Information and'
                         ' Communications Technology', Hemisphere.North,
                         read_hdw_file('ksr')),
              7: _Radar('Kodiak', 'University of Alaska Fairbanks',
                        Hemisphere.North, read_hdw_file('kod')),
              90: _Radar('Longyearbyen', 'University of Centre in Svalbard',
                         Hemisphere.North, read_hdw_file('lyr')),
              9: _Radar('Pykkvibaer', 'University of Leicester',
                        Hemisphere.North, read_hdw_file('pyk')),
              6: _Radar('Prince George', 'University of Saskatchewan',
                        Hemisphere.North, read_hdw_file('pgr')),
              65: _Radar('Rankin Inlet', 'University of Saskatchewan',
                         Hemisphere.North, read_hdw_file('rkn')),
              5: _Radar('Saskatoon', 'University of Saskatchewan',
                        Hemisphere.North, read_hdw_file('sas')),
              2: _Radar('Schefferville', 'CNRS/LPCE', Hemisphere.North,
                        read_hdw_file('sch')),
              8: _Radar('Stokkseyri', 'Lancaster University',
                        Hemisphere.North, read_hdw_file('sto')),
              32: _Radar('Wallops Island', 'JHU Applied Physics Laboratory',
                         Hemisphere.North, read_hdw_file('wal')),
              24: _Radar('Buckland Park', 'La Trobe University',
                         Hemisphere.South, read_hdw_file('bpk')),
              96: _Radar('Dome C',
                         'Institute for Space Astrophysics and Planetology',
                         Hemisphere.South, read_hdw_file('dce')),
              21: _Radar('Falkland Islands', 'British Antarctic Survey',
                         Hemisphere.South, read_hdw_file('fir')),
              4: _Radar('Halley', 'British Antarctic Survey',
                        Hemisphere.South, read_hdw_file('hal')),
              15: _Radar('Kerguelen', 'IRAP/CNRS/IPEV', Hemisphere.South,
                         read_hdw_file('ker')),
              20: _Radar('McMurdo', 'University of Alaska, Fairbanks',
                         Hemisphere.South, read_hdw_file('mcm')),
              11: _Radar('SANAE', 'South African National Space Agency',
                         Hemisphere.South, read_hdw_file('san')),
              22: _Radar('South Pole Station',
                         'University of Alaska, Fairbanks', Hemisphere.South,
                         read_hdw_file('sps')),
              13: _Radar('Syowa East', 'National Institute of Polar Research',
                         Hemisphere.South, read_hdw_file('sye')),
              12: _Radar('Syowa South', 'National Institute of Polar Research',
                         Hemisphere.South, read_hdw_file('sys')),
              14: _Radar('Tiger', 'La Trobe University', Hemisphere.South,
                         read_hdw_file('tig')),
              18: _Radar('Unwin', 'La Trobe University', Hemisphere.South,
                         read_hdw_file('unw')),
              19: _Radar('Zhongshan', 'Polar Research Institute of China',
                         Hemisphere.South, read_hdw_file('zho'))}



# Copyright 2019 SuperDARN Canada, University of Saskatchewan
# Author: Marina Schmidt

"""
This module is to focus on data conversions

All methods in this module are functions for
quick conversions.
"""
from typing import List
import numpy as np
from collections import OrderedDict

from pydarn import DmapArray, DmapScalar

# key is the format char type defined by python,
# item is the DMAP int value for the type
DMAP_FORMAT_TYPES = {'c': 1,  # char = int8 by RST rtypes.h definition
                     'h': 2,  # short integer
                     'i': 3,  # integer (int)
                     'f': 4,  # float
                     'd': 8,  # double
                     's': 9,  # string
                     'q': 10,  # Long int
                     'B': 16,  # unsigned char
                     'H': 17,  # unsigned short int
                     'I': 18,  # unsigned int
                     'Q': 19}  # unsigned long int

# python possible types that could be the python format type characters
# python doesn't have char types, instead it converts them to strings.
# This conversion is okay because RST treats chars as only int8 types.
FORMAT_CONVERSION = {np.int8: 'c',  # RST defines char as an int8 in rtypes.h
                     np.int16: 'h',
                     int: 'i',
                     np.int32: 'i',
                     float: 'f',
                     np.float32: 'f',
                     np.float64: 'd',
                     str: 's',
                     np.str_: 's',
                     np.str: 's',
                     np.int64: 'q',
                     np.uint8: 'B',
                     np.uint16: 'H',
                     np.uint32: 'I',
                     np.uint64: 'Q'}

DMAP_CASTING_TYPES = {'c': np.int8,  # RST defined char
                      'h': np.int16,  # Short
                      'i': int,  # int
                      'f': float,  # Float
                      'd': np.float64,  # Double
                      's': str,  # String
                      'q': np.int64,  # long int
                      'B': np.uint8,  # Unsigned char
                      'H': np.uint16,  # Unsigned short
                      'I': np.uint32,  # Unsigned int
                      'Q': np.uint64}  # Unsigned long int


# !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
#       WARNING: RST treats chars as int8 types,
#       thus single characters are kept are strings
# !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
def dict2dmap(dmap_list: List[dict]) -> List[dict]:
    """
    This method converts list of dictionaries contain keys representing a field
    name for the DMAP format and the item being the value of the field.

    Parameters
    ----------
    dmap_list : List[dict]
        List of dictionaries containing the field names and the
        values of the data

    Return
    ------
    dmap_records : List[dict]
        List of Ordered dictionaries containing dmap data structures
        DmapScalars and DmapArrays.
    """
    dmap_records = []
    for dmap_dict in dmap_list:
        dmap_record = OrderedDict()
        for field, value in dmap_dict.items():
            # if the value is an array then it is a DmapArray
            if isinstance(value, list):
                value = np.array(value)  # keep data structures consistent
            if isinstance(value, np.ndarray):
                shape = list(np.shape(value))
                dimension = np.ndim(value)
                format_type = FORMAT_CONVERSION[value.dtype.type]
                dmap_type = DMAP_FORMAT_TYPES[format_type]
                dmap_data = DmapArray(field, value, dmap_type,
                                      format_type, dimension, shape)

            # else DmapScalar
            else:
                format_type = FORMAT_CONVERSION[type(value)]
                dmap_type = DMAP_FORMAT_TYPES[format_type]
                dmap_data = DmapScalar(field, value, dmap_type, format_type)
            dmap_record[field] = dmap_data
        dmap_records.append(dmap_record)
    return dmap_records


def dmap2dict(dmap_records: List[dict]) -> List[dict]:
    """
    This method converts dmap records containing dmap data structures
    to a list of dictionaries.

    Parameters
    ----------
    dmap_records : List[dict]
        a list of dmap records contain dmap data structures

    Return
    ------
    dmap_dict : List[dict]
        a list of dictionaries containing the name of the fields in the keys
        the data value(s) in the items of the dictionary
    """
    dmap_list = []
    for dmap_record in dmap_records:
        dmap_dict = {field: (data.value if isinstance(data.value, np.ndarray)
                     else DMAP_CASTING_TYPES[data.data_type_fmt](data.value))
                     for field, data in dmap_record.items()}
        dmap_list.append(OrderedDict(dmap_dict))
    return dmap_list



# Copyright 2019 SuperDARN Canada, University of Saskatchewan
# Author: Marci Detwiller
"""
This file contains several exception classes used by the pydarn io borealis
module.

Classes
-------
BorealisFileTypeError
BorealisFieldMissingError
BorealisExtraFieldError
BorealisDataFormatTypeError
BorealisConversionTypesError
BorealisConvert2IqdatError
BorealisConvert2RawacfError
ConvertFileOverWriteError

See Also
--------
BorealisRawacf
BorealisBfiq
BorealisAntennasIq
BorealisRawrf

"""

import logging
pydarn_logger = logging.getLogger('pydarn')


class BorealisFileTypeError(Exception):
    """
    SuperDARN Borealis file type is not implemented or an incorrect type

    Parameters
    ----------
    filename: str
        name of the file associated to the wrong SuperDARN Borealis file type
    file_type: str
        SuperDARN Borealis file type that is not implement or correct

    Attributes
    ----------
    message: str
        The message to display with the error
    """

    def __init__(self, filename: str, file_type: str):
        self.file_type = file_type
        self.filename = filename
        self.message = "{file_type} is not a Borealis file format or has"\
            " not been implemented yet. {filename} was not used. Please check"\
            " the spelling of {file_type}.".format(file_type=self.file_type,
                                                   filename=self.filename)
        Exception.__init__(self, self.message)


class BorealisFieldMissingError(Exception):
    """
    Raised when a required data field is missing.

    Raised when a field from the filetype's fields is not present.

    Parameters
    ----------
    filename: str
        name of the file with the issue
    fields: set
        set of fields that are missing from the file
    record_name: int
        record name of the problem record

    Attributes
    ----------
    message: str
        The message to display with the error
    """

    def __init__(self, filename: str, fields: set, **kwargs):
        if 'record_name' in kwargs.keys():
            self.record_name = kwargs['record_name']
            error_source = 'record {} of file {}'.format(kwargs['record_name'],
                                                         filename)
        else:
            error_source = 'file {}'.format(filename)

        self.filename = filename
        self.fields = fields
        self.message = "The following fields in {source} are missing:"\
            " {fields}".format(source=error_source,
                               fields=self.fields)
        Exception.__init__(self, self.message)


class BorealisExtraFieldError(Exception):
    """
    Raised on extra unknown field in the data.

    Occurs when an extra field is in the data that is not included in
    the Borealis format file type fields

    Parameters
    ----------
    filename: str
        name of the file with the issue
    fields: set
        set of fields that are extra in the file
    record_name: int
        record name of the problem record

    Attributes
    ----------
    message: str
        The message to display with the error
    """

    def __init__(self, filename: str, fields: set, **kwargs):
        if 'record_name' in kwargs.keys():
            self.record_name = kwargs['record_name']
            error_source = 'record {} of file {}'.format(kwargs['record_name'],
                                                         filename)
        else:
            error_source = 'file {}'.format(filename)

        self.fields = fields
        self.message = "The following fields in {source} are not allowed:"\
            " {fields}".format(source=error_source,
                               fields=self.fields)
        Exception.__init__(self, self.message)


class BorealisDataFormatTypeError(Exception):
    """
    Raised on incorrect format of a data field.

    Occurs when a data type does not match the Borealis format file type's
    data types

    Parameters
    ----------
    filename: str
        name of the file with the issue
    incorrect_types: dict
        dict of field : correct type for all fields that were incorrect
    record_name: int
        record name of the problem record

    Attributes
    ----------
    message: str
        The message to display with the error
    """

    def __init__(self, filename: str, incorrect_types: dict, **kwargs):
        if 'record_name' in kwargs.keys():
            self.record_name = kwargs['record_name']
            error_source = 'record {} of file {}'.format(kwargs['record_name'],
                                                         filename)
        else:
            error_source = 'file {}'.format(filename)

        self.incorrect_types = incorrect_types
        self.message = "In {source}, the following parameters"\
            " are not of their given correct type:"\
            " {incorrect}".format(source=error_source,
                                  incorrect=self.incorrect_types)
        Exception.__init__(self, self.message)


class BorealisNumberOfRecordsError(Exception):
    """
    Raised when the file is array structured and does not
    have a consistent number of records across the unshared
    parameters (arrays).

    Parameters
    ----------
    array_types: dict
        dictionary of unshared parameter keys to
        the first dimension of their array (indicating the number
        of records)

    Attributes
    ----------
    message: str
        The message to display with the error

    See Also
    --------
    restructure_borealis.py
    """

    def __init__(self, filename: str, array_types: dict):
        self.filename = filename
        self.array_types = array_types
        self.message = "The number of records in file {filename} cannot "\
            "be determined due to varying shapes of arrays for the unshared "\
            "parameters. All arrays of the unshared parameters should "\
            "have the same first dimension size equal to number of records: "\
            "{array_types}".format(filename=self.filename,
                                   array_types=self.array_types)
        Exception.__init__(self, self.message)


class BorealisConversionTypesError(Exception):
    """
    SuperDARN Borealis filetype cannot be converted to desired type.

    Parameters
    ----------
    filename: str
        name of the file attempting to convert
    origin_filetype: str
        Borealis origin filetype, ex. bfiq.
    dmap_filetype: str
        desired SuperDARN DMap type, ex. iqdat.

    Attributes
    ----------
    message: str
        The message to display with the error
    """

    def __init__(self, dmap_filename: str, origin_filetype: str,
                 allowed_types: dict):
        self.message = "Records destined to be converted and written to "\
                       "{dmap_filename} cannot be converted from origin "\
                       "filetype {origin_filetype} to a dmap "\
                       "filetype because origin filetype does not map "\
                       "to any currently available dmap filetypes: "\
                       "{allowed_types}".format(dmap_filename=dmap_filename,
                                                origin_filetype=origin_filetype,
                                                allowed_types=allowed_types)
        Exception.__init__(self, self.message)


class BorealisConvert2IqdatError(Exception):
    """
    Raised when the file cannot be converted to SDARN DMap iqdat format.

    Parameters
    ----------
    error_str: str
        explanation for why the file cannot be converted to SDARN iqdat.

    Attributes
    ----------
    message: str
        The message to display with the error
    """

    def __init__(self, error_str: str):
        self.message = "The file cannot be converted to SDARN DMap iqdat due"\
            " to the following error: {error_str}"\
            "".format(error_str=error_str)
        Exception.__init__(self, self.message)


class BorealisConvert2RawacfError(Exception):
    """
    Raised when the file cannot be converted to SDARN DMap rawacf format.

    Parameters
    ----------
    error_str: str
        explanation for why the file cannot be converted to SDARN rawacf.

    Attributes
    ----------
    message: str
        The message to display with the error
    """

    def __init__(self, error_str: str):
        self.message = "The file cannot be converted to SDARN DMap rawacf "\
            "due to the following error: {error_str}"\
            "".format(error_str=error_str)
        Exception.__init__(self, self.message)


class BorealisRestructureError(Exception):
    """
    Raised when the file cannot be restructured to or from site/array
    styles.

    Parameters
    ----------
    error_str: str
        explanation for why the file cannot be restructured.

    Attributes
    ----------
    message: str
        The message to display with the error

    See Also
    --------
    restructure_borealis.py
    """

    def __init__(self, error_str: str):
        self.message = "The file cannot be restructured due to the "\
            " following error: {error_str}"\
            "".format(error_str=error_str)
        Exception.__init__(self, self.message)


class BorealisStructureError(Exception):
    """
    Raised when the file has major structural issues and may not
    be the correct style type (arrays vs records).

    Parameters
    ----------
    error_str: str
        explanation for why the file cannot be restructured.

    Attributes
    ----------
    message: str
        The message to display with the error

    See Also
    --------
    restructure_borealis.py
    """

    def __init__(self, error_str: str):
        self.message = "Structural errors found. You may be"\
                " attempting to use the file as the wrong"\
                " structure ('site' or 'array'"\
                " supported). Please check if you are interested"\
                " in records or arrays."\
                "{error_str}".format(error_str=error_str)
        Exception.__init__(self, self.message)


class ConvertFileOverWriteError(Exception):
    """
    Trying to write to a file that also reading.

    Parameters
    ----------
    filename: str
        name of the file trying to both read and write from/to

    Attributes
    ----------
    message: str
        The message to display with the error
    """

    def __init__(self, filename: str):
        self.message = "Writing to {filename} not permitted while reading"\
            " the file as source.".format(filename=filename)
        Exception.__init__(self, self.message)



# Copyright 2019 SuperDARN Canada, University of Saskatchewan
# Author: Marci Detwiller
"""
This file contains classes and functions for
converting of Borealis file types.

Classes
-------
BorealisRead: Uses BorealisSiteRead and BorealisArrayRead
    to read a Borealis HDF5 file
BorealisWrite: Uses BorealisSiteWrite and BorealisArrayWrite
    to write a Borealis HDF5 file

Exceptions
----------
BorealisFileTypeError
BorealisFieldMissingError
BorealisExtraFieldError
BorealisDataFormatTypeError
BorealisNumberOfRecordsError
BorealisConversionTypesError
BorealisConvert2IqdatError
BorealisConvert2RawacfError
BorealisRestructureError
BorealisStructureError
ConvertFileOverWriteError

See Also
--------
BorealisSiteRead
BorealisSiteWrite
BorealisArrayRead
BorealisArrayWrite
BorealisConvert

For more information on Borealis data files and how they convert to SDARN
filetypes, see: https://borealis.readthedocs.io/en/latest/
"""

import logging

from collections import OrderedDict
from typing import Union

from pydarn import borealis_exceptions
from .borealis_site import BorealisSiteRead, BorealisSiteWrite
from .borealis_array import BorealisArrayRead, BorealisArrayWrite

pydarn_log = logging.getLogger('pydarn')


class BorealisRead():
    """
    Class for reading Borealis filetypes of both array and site
    structure. This class abstracts and redirects the file being read to
    the correct class (BorealisSiteRead or BorealisArrayRead).

    See Also
    --------
    BorealisSiteRead
    BorealisArrayRead
    BorealisRawacf
    BorealisBfiq
    BorealisAntennasIq
    BorealisRawrf

    Attributes
    ----------
    filename: str
        The filename of the Borealis HDF5 file being read.
    borealis_filetype: str
        The type of Borealis file. Types include:
        'bfiq'
        'antennas_iq'
        'rawacf'
        'rawrf'
    reader: Union[BorealisSiteWrite, BorealisArrayWrite]
        the wrapped BorealisSiteRead or BorealisArrayRead instance
    borealis_file_structure: Union[str, None]
        The structure of the data. 'site' structure means record-by-record
        style, as is written by the Borealis radar on-site. 'array'
        structure means a restructured file where parameters
        that are consistent throughout the whole file are provided
        once and data is restructured into arrays where the
        first dimension is num_records. 'array' structure is
        intended for downstream users to more easily read and use
        the data.
    records: dict
    arrays: dict
    record_names: list[str]
    """

    def __init__(self, filename: str, borealis_filetype: str,
                 borealis_file_structure: Union[str, None] = None):
        """
        Reads Borealis array file types into a dictionary.

        Parameters
        ----------
        filename: str
            file name containing Borealis hdf5 data.
        borealis_filetype: str
            The type of Borealis file. Types include:
            'bfiq'
            'antennas_iq'
            'rawacf'
            'rawrf'
        borealis_file_structure: Union[str, None]
            The write structure of the file provided. Possible types are
            'site', 'array', or None. If None (default), array will be
            attempted first followed by site.

        Raises
        ------
        OSError
            Unable to open file
        BorealisStructureError
            Unknown structure type.
        """
        self.filename = filename

        if borealis_filetype not in ['bfiq', 'antennas_iq', 'rawacf', 'rawrf']:
            raise borealis_exceptions.BorealisFileTypeError(
                self.filename, borealis_filetype)
        self.borealis_filetype = borealis_filetype

        if borealis_file_structure is None:
            self._reader, self._borealis_file_structure = \
                self.return_reader(self.filename, self.borealis_filetype)
        elif borealis_file_structure == 'site':
            self._reader = BorealisSiteRead(self.filename,
                                            self.borealis_filetype)
            self._borealis_file_structure = 'site'
        elif borealis_file_structure == 'array':
            self._reader = BorealisArrayRead(self.filename,
                                             self.borealis_filetype)
            self._borealis_file_structure = 'array'
        else:  # unknown structure
            raise borealis_exceptions.BorealisStructureError("Unknown"
                                                             " structure"
                                                             " type: {}"
                                                             "".format(borealis_file_structure))

    def __repr__(self):
        """ for representation of the class object"""
        # __class__.__name__ allows to grab the class name such that
        # when a class inherits this one, the class name will be the child
        # class and not the parent class
        return "{class_name}({filename}, {borealis_filetype})"\
            "".format(class_name=self.__class__.__name__,
                      filename=self.filename,
                      borealis_filetype=self.borealis_filetype)

    def __str__(self):
        """ for printing of the class object"""

        return "Reading from {filename}"\
            "".format(filename=self.filename)

    @property
    def borealis_file_structure(self):
        """
        The structure of the file read, 'site' or 'array'. Default None.
        """
        return self._borealis_file_structure

    @property
    def record_names(self):
        """
        A sorted list of the set of record names in the HDF5 file read.
        These correspond to Borealis file record write times (in ms since
        epoch), and are equal to the group names in the site file types.
        """
        return self._reader.record_names

    @property
    def records(self):
        """
        The Borealis data in a dictionary of records, according to the
        site file format.
        """
        return self._reader.records

    @property
    def arrays(self):
        """
        The Borealis data in a dictionary of arrays, according to the
        restructured array file format.
        """
        return self._reader.arrays

    @staticmethod
    def return_reader(borealis_hdf5_file: str, borealis_filetype: str) -> \
                     (Union[BorealisArrayRead, BorealisSiteRead], str):
        """
        Attempts to read a file as array and then as site. Returns if
        any read is successful.

        Parameters
        ----------
        borealis_hdf5_file
            Borealis file to read. Either array or site style.
        borealis_filetype
            Borealis filetype. Possible types include antennas_iq,
            bfiq, rawacf.

        Returns
        -------
        reader
            instance of BorealisArrayRead or BorealisSiteRead, depending on
            reader success
        reader_type
            'array' or 'site', depending on which read was successful

        Raises
        ------
        BorealisExtraFieldError
        BorealisMissingFieldError
        BorealisDataFormatTypeError

        See Also
        --------
        BorealisArrayRead
        BorealisSiteRead
        """
        try:
            reader = BorealisArrayRead(borealis_hdf5_file, borealis_filetype)
            return reader, 'array'
        except (borealis_exceptions.BorealisExtraFieldError,
                borealis_exceptions.BorealisFieldMissingError,
                borealis_exceptions.BorealisStructureError):
            pydarn_log.info('{} is not array restructured. Attempting site'
                            ' read.'.format(borealis_hdf5_file))
            try:
                reader = BorealisSiteRead(borealis_hdf5_file,
                                          borealis_filetype)
                return reader, 'site'
            except Exception as err:
                pydarn_log.info('{} is not site structured. Raising reader'
                                ' errors.'.format(borealis_hdf5_file))
                raise err


class BorealisWrite():
    """
    Class for writing Borealis filetypes of both array and site
    structure. This class abstracts and redirects the file write to
    the correct class (BorealisSiteWrite or BorealisArrayWrite).

    See Also
    --------
    BorealisSiteWrite
    BorealisArrayWrite
    BorealisRawacf
    BorealisBfiq
    BorealisAntennasIq
    BorealisRawrf

    Attributes
    ----------
    filename: str
        The filename of the Borealis HDF5 file being written.
    borealis_filetype: str
        The type of Borealis file. Types include:
        'bfiq'
        'antennas_iq'
        'rawacf'
        'rawrf'
    writer: Union[BorealisSiteWrite, BorealisArrayWrite]
        the wrapped BorealisSiteWrite or BorealisArrayWrite instance
    borealis_file_structure: Union[str, None]
        The structure of the data. 'site' structure means record-by-record
        style, as is written by the Borealis radar on-site. 'array'
        structure means a restructured file where parameters
        that are consistent throughout the whole file are provided
        once and data is restructured into arrays where the
        first dimension is num_records. 'array' structure is
        intended for downstream users to more easily read and use
        the data.
    compression: Union[str, None]
        The type of compression the file is written with. Default is
        None for site structure, and 'zlib' for array structure.
        Site structured files will have no compression to match
        files actually written on-site, where records are constantly
        appended so there is no compression. Array structured files
        will have 'zlib' compression to enable a faster read time
        for downstream users.
    arrays: dict
        The Borealis data in a dictionary of arrays, according to the
        restructured array file format.
    """

    def __init__(self, filename: str, borealis_data: Union[dict, OrderedDict],
                 borealis_filetype: str,
                 borealis_file_structure: Union[str, None] = None,
                 **kwargs):
        """
        Write borealis records to a file.

        Parameters
        ----------
        filename: str
            Name of the file the user wants to write to
        data: Union[dict, OrderedDict]
            borealis data dictionary. Can be arrays or records.
        borealis_filetype: str
            The type of Borealis file. Restructured types include:
            'bfiq'
            'antennas_iq'
            'rawacf'
        borealis_file_structure: Union[str, None]
            The structure of the data provided. Possible types are
            'site', 'array', or None. If None (default), array write will be
            attempted first followed by site.
        hdf5_compression: str
            A kwarg key name, giving a string representing compression type.
        """
        self.filename = filename
        self.data = borealis_data
        self.borealis_filetype = borealis_filetype

        if borealis_file_structure is None:
            self._writer, self._borealis_file_structure = \
                self.return_writer(self.filename, self.data,
                                   self.borealis_filetype, **kwargs)
        elif borealis_file_structure == 'site':
            self._writer = BorealisSiteWrite(self.filename, self.data,
                                             self.borealis_filetype,
                                             **kwargs)
            self._borealis_file_structure = 'site'
        elif borealis_file_structure == 'array':
            self._writer = BorealisArrayWrite(self.filename, self.data,
                                              self.borealis_filetype, **kwargs)
            self._borealis_file_structure = 'array'
        else:  # unknown structure
            raise borealis_exceptions.BorealisStructureError('Unknown '
                                                             'structure '
                                                             'type: {}'.format(borealis_file_structure))

    def __repr__(self):
        """For representation of the class object"""

        return "{class_name}({filename}, {current_record_name})"\
               "".format(class_name=self.__class__.__name__,
                         filename=self.filename,
                         current_record_name=self.current_record_name)

    def __str__(self):
        """For printing of the class object"""

        return "Writing to filename: {filename} at record name: "\
               "{current_record_name}".format(filename=self.filename,
                                              current_record_name=self.current_record_name)

    @property
    def borealis_file_structure(self):
        """
        The structure of the file read, 'site' or 'array'. Default None.
        """
        return self._borealis_file_structure

    @property
    def compression(self):
        """
        The type of compression the file is written with. Default is
        None for site structure, and 'zlib' for array structure.
        Site structured files will have no compression to match
        files actually written on-site, where records are constantly
        appended so there is no compression. Array structured files
        will have 'zlib' compression to enable a faster read time
        for downstream users.
        """
        return self._writer.compression

    @property
    def record_names(self):
        """
        A sorted list of the set of record names in the HDF5 file read.
        These correspond to Borealis file record write times (in ms since
        epoch), and are equal to the group names in the site file types.
        """
        return self._writer.record_names

    @property
    def records(self):
        """
        The Borealis data in a dictionary of records, according to the
        site file format.
        """
        return self._writer.records

    @property
    def arrays(self):
        """
        The Borealis data in a dictionary of arrays, according to the
        restructured array file format.
        """
        return self._writer.arrays

    @staticmethod
    def return_writer(filename: str, data: Union[dict, OrderedDict],
                      borealis_filetype: str, **kwargs) -> \
            (Union[BorealisArrayRead, BorealisSiteRead], str):
        """
        Attempts to write a file as array and then site. Returns if any writer
        is successful.

        Parameters
        ----------
        filename
            Filename to write to. Either array or site style.
        data
            Data to write out. Can be arrays dictionary or OrderedDict of
            records.
        borealis_filetype
            Borealis filetype. Possible types include antennas_iq,
            bfiq, rawacf, rawrf (site only).
        kwargs:
            keyword arguments for writer call. Only supported is
            'hdf5_compression'

        Returns
        -------
        writer
            instance of BorealisArrayWrite or BorealisSiteWrite, depending on
            writer success
        writer_type
            'array' or 'site', depending on which write was successful

        Raises
        ------
        BorealisExtraFieldError
        BorealisMissingFieldError
        BorealisDataFormatTypeError

        See Also
        --------
        BorealisArrayWrite
        BorealisSiteWrite
        """
        try:
            writer = BorealisArrayWrite(filename, data,
                                        borealis_filetype, **kwargs)
            return writer, 'array'
        except (borealis_exceptions.BorealisExtraFieldError,
                borealis_exceptions.BorealisFieldMissingError,
                borealis_exceptions.BorealisStructureError):
            pydarn_log.info('Data provided is not array restructured. '
                            'Attempting site write to file {}.'
                            ''.format(filename))
            try:
                writer = BorealisSiteWrite(filename, data,
                                           borealis_filetype, **kwargs)
                return writer, 'site'
            except Exception as err:
                pydarn_log.info('Data provided is not site structured. '
                                'Raising writer error.')
                raise err
